/*
 *  linux/fs/hfs/btree.c
 *
 *  Copyright (C) 1995  Paul H. Hargrove
 *  This file may be distributed under the terms of the GPL
 *
 *  This file contains the code neccesary to manipulate the B-tree
 *  structure, which is used extensively in HFS.
 *  The catalog file and the extens file are each a B-tree.
 *
 */

#include <linux/config.h>
#ifdef MODULE
#include <linux/module.h>
#endif

#include "hfs_fs.h"
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/sched.h>
#include <linux/fs.h>

/* some defines to keep the code from getting too long-winded */
#define REC_TBL(brec)	((HFS_WORD *)((brec)->bnode->bh->b_data+HFS_SECTOR_SIZE))
#define HEAD(brec)	((struct node_head *)(brec)->bnode->bh->b_data)
#define KEY(brec, n)	((struct hfs_rec_key *)((brec)->bnode->bh->b_data + from_w((brec)->record_tbl[-(n)])))
#define RECORD(key)	((void *)(key) + (((key)->KeyLen + 2) & ~1))

/*======== Private Data Types ========*/

/*
 * The btree header record
 *
 * This data structure is in the first node (512-byte block) of each
 * B-tree file.
 */
struct tree_head {
  /* The number of levels in the tree */
	HFS_WORD	tree_depth	DONT_ALIGN;
  /* The node number of the root of the tree */
	HFS_LWORD	root_node	DONT_ALIGN;
  /* The number of leaf nodes */
	HFS_LWORD	leaf_recs	DONT_ALIGN;
  /* The number of the first leaf node */
	HFS_LWORD	first_leaf	DONT_ALIGN;
  /* The number of the last leaf node */
	HFS_LWORD	last_leaf	DONT_ALIGN;
  /* The number of bytes in a node.  Currently, only 512 is allowed */
	HFS_WORD	node_size	DONT_ALIGN;
  /* The maximum length of a "key" in this B-tree */
  /* Keys in internal nodes are all of this length. */
  /* Keys in leaf nodes are of variable length. */
	HFS_WORD	key_len		DONT_ALIGN;
  /* The total number of nodes currently in the file */
	HFS_LWORD	total_nodes	DONT_ALIGN;
  /* The number of nodes in the file, but not used in the tree */
	HFS_LWORD	free_nodes	DONT_ALIGN;
};

/*
 * A btree node header
 *
 * This structure begins each node in the B-tree file.
 */
struct node_head {
  /* Forward and Backward "pointers" (node numbers), forming */
  /* a linked-list of all nodes at the same level. */
	HFS_LWORD	ndFLink		DONT_ALIGN;
	HFS_LWORD	ndBLink		DONT_ALIGN;
  /* Type of node (see eunm nd_type, below) */
	HFS_BYTE	ndType		DONT_ALIGN;
  /* The level of this node in the B-tree, where leaves are level 1 */
	HFS_BYTE	ndLevel		DONT_ALIGN;
  /* Number of records in this node of the B-tree */
	HFS_WORD	ndNRecs		DONT_ALIGN;
  /* Reseved field / Padding */
	HFS_WORD	ndReserv	DONT_ALIGN;
};

/* 
 * Legal values of node_nead.ndType
 */
enum nd_type {
	ND_INDX = 0x00,		/* An internal (index) node */
	ND_HDR  = 0x01,		/* The tree header node (node 0) */
	ND_MAP  = 0x02,		/* Holds part of the bitmap of used nodes */
	ND_LEAF = 0xFF,		/* A leaf (ndLevel==1) node */
};

/*
 * A in-core B-tree.
 *
 * This structure holds information from the tree_head, MDB
 * (superblock) and other information needed to work with the B-tree.
 * 
 * Both the superblock and the tree_head are kept in-core, so this
 * information is mostly redundant.  However, on a little-endian
 * processor like the x86 it is convenient to have it all in the
 * proper byte-order.  It is also nice to have it all in one place.
 */
struct hfs_btree {
  /* A magic number */
	int			magic;
  /* The function used to compare two keys in this B-tree */
  /* returns 0, positive or negative, like strcmp() */
	int			(*compf)(struct hfs_rec_key *, struct hfs_rec_key *);
  /* The hfs_file structure giving the location of the B-tree on disk */
	struct hfs_file		file;
  /* Pointer to the in-core copy of node 0 */
	struct hfs_bnode	*head;
  /* Pointer to the in-core copy of the root node */
	struct hfs_bnode	*root;
  /* The device containing this file-system */
	dev_t			dev;
  /* The number of bytes in the B-tree file */
	unsigned long		tree_size;
  /* Fields from the tree_head: */
	unsigned int		root_node;
	unsigned int		leaf_recs;
	unsigned int		first_leaf;
	unsigned int		last_leaf;
	unsigned int		total_nodes;
	unsigned int		free_nodes;
	unsigned short		tree_depth;
	unsigned short		key_len;
};

/*
 * An in-core B-tree node
 *
 * This structure holds information from the node_head in
 * native byte-order, a pointer to the buffer_head which
 * contains the actual node, and fields neccesary for locking
 * access to the node durring updates.
 */
struct hfs_bnode {
  /* A magic number */
	int			magic;
  /* The buffer_head containing the a actual node */
	struct buffer_head	*bh;
  /* The tree to which this node belongs */
	struct hfs_btree	*tree;
  /* Next and Previous pointers for a linked-list of */
  /* in-core nodes in the same slot in the hash table */
	struct hfs_bnode	*prev;
	struct hfs_bnode	*next;
  /* A process wishing to modify this node must wait on this queue */
  /* until it is the only process with access to the node. */
  /* No new process may gain access to this node while this queue is non-empty. */
	struct wait_queue	*wqueue;
  /* If a process cannot gain access to this node because wqueue */
  /* is non-empty, then it waits on this queue. */
  /* A process will also wait on this queue if it wishes to set resrv */
  /* when it is all ready set. */
	struct wait_queue	*rqueue;
  /* The number of processes with access to this node */
	int			count;
  /* Boolean indicating that one of the current users of this node */
  /* may potentially modify this node before it is done with it. */
  /* While set no other potential modifier may gain access. */
	int			resrv;
  /* Boolean indicating the current user of this is the only user, */
  /* and is in the process of modifying the node. */
	int			lock;
  /* Boolean indicating that this node is to be kept in-core even when count==0. */
	int			sticky;
  /* Three fields used to locate the correct node. */
 	dev_t			dev;
	int			fid;
	int			node;
};

/*
 * Access types used when requesting access to a B-node
 */
enum locktype {
	LOCK_NONE = 0,		/* Illegal */
	LOCK_READ,		/* read-only access */
	LOCK_RESRV,		/* might potentially modify */
	LOCK_WRITE,		/* will modify now */
};

/* 
 * flags (or'ed together) to bfind()
 */
/* request LOCK_READ access to the nodes, otherwise LOCK_RESRV */
#define	BFIND_RDONLY	0x01
/* keep access to each "required" node from the root to the matching leaf */
#define	BFIND_CHAIN	0x02
/* only accept an exect match, otherwise take the largest node less than the key */
#define	BFIND_EXACT	0x04
/* define "required" for BFIND_CHAIN to mean that access to an internal node is */
/* kept as long as insertion of a record in a child could potentialy modify it. */
/* XXX NOT IMPLEMENTED */
#define	BFIND_INSERT	0x08
/* define "required" for BFIND_CHAIN to mean that access to an internal node is */
/* kept as long as deletion of a record in a child could potentialy modify it. */
/* XXX NOT IMPLEMENTED */
#define	BFIND_DELETE	0x10
/* request LOCK_WRITE access to the matching node when it is located */
#define	BFIND_LOCK	0x20

/*======== Private Internal Functions ========*/

#define CACHELEN	15	/* use (2**n)-1 for some n>2 */

/* a hash-table based cache of in-core bnodes */
static struct hfs_bnode *bnode_cache[CACHELEN];

/* a simple hash function */
static __inline__ const int bhashfn(dev_t dev, int fid, int node) {
	return (dev ^ fid ^ node) % CACHELEN;
}

/* macro to access the hash table */
#define bhash(dev, fid, node)	(bnode_cache[bhashfn(dev, fid, node)])

/* debugging statistics */
#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
static int hit = 0;
static int miss = 0;
static int count = 0;
#endif

/*
 * This function is called to remove a bnode from the cache.
 * It also release the buffer_head, and the kmalloc'd bnode strucure.
 */
static void bnode_delete(struct hfs_bnode *bn) {
/* NOTE: all callers check that bn!=NULL */
#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
	--count;
#endif
	if (bn->next)
		bn->next->prev = bn->prev;
	if (bn->prev)
		bn->prev->next = bn->next;
	if (bhash(bn->dev, bn->fid, bn->node) == bn)
		bhash(bn->dev, bn->fid, bn->node) = bn->next;
	if (bn->bh)
		brelse(bn->bh);
	kfree_s(bn, sizeof(struct hfs_bnode));
}

/*
 * This function is called when a process is done using a
 * bnode.  If the proper conditions are met, then we remove it
 * from the cache.
 */
static void bnode_relse(struct hfs_bnode *bn, int lock_type)
{
/* NOTE: all callers check that bn!=NULL */
	if (lock_type == LOCK_WRITE) {
		mark_buffer_dirty(bn->bh, 1);
	}
	/*
	 * we update the lock state of the node if any of the following are true:
	 *  1) it is still in use
	 *  2) it is the only entry in the cache slot
	 *  3) it is "sticky" (such as the b-tree head and root)
	 *
	 * Otherwise we just delete it.
	 */
	if ((--bn->count) || (!bn->next && !bn->prev) || (bn->sticky)) {
	  
		switch (lock_type) {
		case LOCK_READ:
			if ((bn->count == 1) && bn->wqueue)
				wake_up(&bn->wqueue);
			break;

		case LOCK_WRITE:
			bn->lock = 0;
		case LOCK_RESRV:
			bn->resrv = 0;
			wake_up(&bn->rqueue);
			break;

		default:
			printk("hfs_brec_relse: invalid lock type: %d\n", lock_type);
		}

		/* If we are keeping the bnode only because it is the only entry */
		/* in its cache slot (in other words we have nothing better to do */
		/* with its cache slot), then we release the buffer_head. */
		if (!bn->count && !bn->sticky) {
			brelse(bn->bh);
			bn->bh = NULL;
		}
	} else {
		bnode_delete(bn);
	}
}

/* 
 * This function creates a bnode, and inserts it in the cache.
 * The newly created bnode has count=resrv=lock=0 and bh=NULL.
 * The caller must set these fields if necessary.
 */
static __inline__ struct hfs_bnode * bnode_create(struct hfs_btree *tree, int node)
{
/* NOTE: all callers check that tree!=NULL */
	struct hfs_bnode *tmp, *bnode;

	if ((bnode=kmalloc(sizeof(struct hfs_bnode), GFP_KERNEL))) {
#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
		++count;
#endif
		/* initialize the structure */
		memset(bnode, 0, sizeof(struct hfs_bnode));
		bnode->magic = HFS_BNODE_MAGIC;
		bnode->tree = tree;
		bnode->dev = tree->dev;
		bnode->fid = tree->file.fid;
		bnode->node = node;

		/* insert it in the cache */
		bnode->next = tmp = bhash(bnode->dev, bnode->fid, node);
		if (tmp)
			tmp->prev = bnode;
		bnode->prev = NULL;
		bhash(bnode->dev, bnode->fid, node) = bnode;
		return bnode;
	}
	return NULL;
}

/*
 * This is a boolean function that returns false if the bnode
 * matches the supplied dev, fid and node.
 */
static __inline__ int not_match(const struct hfs_bnode *bn, dev_t dev, int fid, int node)
{
	return (bn->dev != dev) || (bn->fid != fid) || (bn->node != node);
} 

/*
 * This function does the INITIAL locking of a bnode.
 * LOCK_WRITE is not legal, since a LOCK_RESRV lock must
 * be obtained before it is converted to a LOCK_WRITE lock.
 * This lock-conversion is done by brec_lock().
 */
static void bnode_lock(struct hfs_bnode *bn, int lock_type)
{
/* NOTE: all callers check that bn!=NULL */
	++bn->count;
	switch (lock_type) {
	case LOCK_READ:
		while (bn->lock || bn->wqueue)
			sleep_on(&bn->rqueue);
		break;

	case LOCK_RESRV:
		while (bn->resrv)
			sleep_on(&bn->rqueue);
		++bn->resrv;
		break;

	default:
		printk("hfs_bnode_lock: invalid lock type\n");
	}
}

/*
 * This function is called with lock_type=LOCK_READ or LOCK_RESERV to
 * obtain access to a bnode.
 * The cache is searched for the node.  If it not found there it is
 * created by bnode_create() and then read from disk.
 */
static struct hfs_bnode * read_node(struct hfs_btree *tree, int node, int lock_type)
{
/* NOTE: all callers check that tree!=NULL */
	struct hfs_bnode *bn;
	dev_t dev = tree->dev;
	int fid = tree->file.fid;

	if ((lock_type != LOCK_READ) &&
	    (lock_type != LOCK_RESRV)) {
		printk("hfs_read_node: invalid lock type: %d\n", lock_type);
		return NULL;
	}

#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
	printk("read_node: %c %d:%d ", ((lock_type==LOCK_READ)?'R':'V'), tree->file.fid, node);
#endif

	/* look for the node in the cache. */
	bn = bhash(dev, fid, node);
	if (bn && not_match(bn, dev, fid, node)) {
		if (bn->magic != HFS_BNODE_MAGIC) {
			printk("hfs_read_node: bnode cache is corrupted.\n");
			return NULL;
		} else if (!bn->count) {
			/* first is unused so discard it */
			bnode_delete(bn);
			bn = NULL;
		} else {
			/* run through whatever else is in the list */
			while (bn && not_match(bn, dev, fid, node))
				bn = bn->next;
		}
	}

#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
	if (bn)
		++hit;
	else
		++miss;
	printk("%c (%d%%) ", ((bn)?'+':'-'), ((hit*100)/(hit+miss)));
#endif

	if (bn || (bn=bnode_create(tree, node))) {
		if (!bn->count && !bn->bh) {
			bn->count = bn->resrv = bn->lock = 1;
			bn->bh = bread(dev, hfs_block_map(&tree->file, node), HFS_SECTOR_SIZE);
			--bn->count;
			bn->resrv = bn->lock = 0;
			wake_up(&bn->rqueue);
			if (!bn->bh) {
				if (!bn->count)
					bnode_delete(bn);
				printk("hfs_read_node: unable to read node from dev 0x%04x\n", dev);
				return NULL;
			}
		}
		bnode_lock(bn, lock_type);
	}

#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
	if (!bn)
		printk("failed\n");
	else
		printk("use %d(%d) lvl %d [%d]\n", bn->count, bn->bh->b_count,
			((struct node_head *)(bn->bh->b_data))->ndLevel, count);
#endif

	return bn;
}

/*
 * This is the comparison function used for the catalog b-tree.
 * It returns <0 if key1<key2, >0 if key1>key2, and 0 if key1==key2
 */
static int cat_compare(struct hfs_rec_key *key1, struct hfs_rec_key *key2)
{
/* NOTE: all callers check key1,key2 != NULL */
	int parents;

	parents = from_l(key1->u.ckr.ParID) - from_l(key2->u.ckr.ParID);
	if (parents != 0)
		return parents;
	return hfs_strcmp(key1->u.ckr.CName, key2->u.ckr.CName);
}
/*
 * This is the comparison function used for the catalog b-tree.
 * It returns <0 if key1<key2, >0 if key1>key2, and 0 if key1==key2
 */

/*
 * This is the comparison function used for the extents b-tree.
 * It returns <0 if key1<key2, >0 if key1>key2, and 0 if key1==key2
 */
static int ext_compare(struct hfs_rec_key *key1, struct hfs_rec_key *key2)
{
/* NOTE: all callers check key1,key2 != NULL */
	int tmp;

	tmp = (int)from_l(key1->u.xkr.FNum) - (int)from_l(key2->u.xkr.FNum);
	if (tmp != 0)
		return tmp;
	tmp = (int)key1->u.xkr.FkType - (int)key2->u.xkr.FkType;
	if (tmp != 0)
		return tmp;
	return (int)from_w(key1->u.xkr.FABN) - (int)from_w(key2->u.xkr.FABN);
}

/*
 * This function allocates an hfs_brec structure.
 * This is the structure used to comunicate with the catalog and
 * extents codes about records in b-trees.
 */
static __inline__ struct hfs_brec *alloc_brec(void) {
	struct hfs_brec *brec;
#if defined(DEBUG_BRECS) || defined(DEBUG_ALL)
	extern int hfs_bcount;
	++hfs_bcount;
#endif
	if ((brec=kmalloc(sizeof(struct hfs_brec), GFP_KERNEL))) {
		memset(brec, 0, sizeof(struct hfs_brec));
		brec->magic = HFS_BREC_MAGIC;
	}
	return brec;
}

/*
 * This function obtains LOCK_WRITE access to the bnode containing
 * this hfs_brec.  All decendents in the path from this record to
 * the leaf are given LOCK_WRITE access, and all ancestors in the
 * path from the root to here are released.
 */
static void brec_lock(struct hfs_brec *brec) {
/* NOTE: all callers check brec!=NULL */
	if (brec->lock_type == LOCK_RESRV) {
		hfs_brec_relse(&brec->up);
		do {
			while (brec->bnode->count > 1)
				sleep_on(&brec->bnode->wqueue);
			++brec->bnode->lock;
			brec->lock_type = LOCK_WRITE;
		} while ((brec = brec->down));
	} else {
		printk("hfs_brec_lock: record not reserved\n");
	}
}

/*
 * Obtain access to the root node of the tree.
 * One must simultaniously get access to the tree_head, since
 * it is the "parent of root".  In other words, if a modification
 * to the root changes the depth of the tree, a couple fields in the
 * tree_head must be changed too.
 * Returns a partially initialized hfs_brec for the root node.
 *
 * Called only by bfind().
 */
static __inline__ struct hfs_brec *get_root(struct hfs_btree *tree, int flags)
{
	struct hfs_brec *head, *root;
	int lock_type;
	
	lock_type = (flags & BFIND_RDONLY) ? LOCK_READ : LOCK_RESRV;

	if (!(head = alloc_brec()))
		goto bail2;
	head->bnode = tree->head;
	head->lock_type = lock_type;
	bnode_lock(head->bnode, lock_type);

	if (!(root = alloc_brec()))
		goto bail1;
	root->bnode = tree->root;
	root->lock_type = lock_type;
	bnode_lock(root->bnode, lock_type);

	if (flags & BFIND_CHAIN) {
		root->up = head;
		head->down = root;
	} else {
		hfs_brec_relse(&head);
	}

	return root;

bail1:
	hfs_brec_relse(&head);
bail2:
	return NULL;
}

/*
 * Called with brec pointing to a record in an internal node.
 * Returns a pointer to a partially initialized hfs_brec for the child node.
 *
 * Called only by bfind().
 *
 * XXX wastes time/space by allocating a new brec even when not needed
 */
static __inline__ struct hfs_brec *get_child(struct hfs_brec *brec, int flags)
{
	struct hfs_brec *new;

	if ((new = alloc_brec())) {
		/* read in the node */
		if ((new->bnode = read_node(brec->bnode->tree,
					    from_l(*brec->data.child),
					    brec->lock_type))) {
			new->lock_type = brec->lock_type;
			if (flags & BFIND_CHAIN) {
				new->up = brec;
				brec->down = new;
			} else {
				hfs_brec_relse(&brec);
			}
			return new;
		}
		hfs_brec_relse(&new);
	}
	hfs_brec_relse(&brec);
	return NULL;
}

/*
 * This function is very important.
 * It has sole responsibility for locating existing nodes in a b-tree.
 * It's actions are defined by the value of the flags argument.
 * 
 * It returns 0 on success or an error code.
 * **brec is the located record.
 *
 * XXX BFIND_INSERT and BFIND_DELETE are not yet implemented. 
 * XXX This effects only efficiency, not correctness.
 */
static int bfind(struct hfs_btree *tree, struct hfs_rec_key *target, struct hfs_brec **brec, int flags)
{
	int result, ntype;
	struct hfs_brec *tmp;

	if (!brec)
		return -EINVAL;

	*brec = NULL;

	if (!tree || !target)
		return -EINVAL;

	if (!(tmp = get_root(tree, flags)))
		return -ENOMEM;

	do {
		tmp->record_tbl = REC_TBL(tmp);
		tmp->record = from_w(HEAD(tmp)->ndNRecs);

		/* reverse linear search yielding largest key <= target */
		do {
			tmp->key = KEY(tmp, tmp->record);
			result = (tree->compf)(target, tmp->key);
		} while ((result<0) && (--tmp->record));

 		/* see if all keys > target */
		if (!tmp->record) {
			hfs_brec_relse(&tmp);
			return -ENOENT;
		}

		tmp->data.generic = RECORD(tmp->key);
		ntype = HEAD(tmp)->ndType;

		/* get next node if necessary */
		if ((ntype == ND_INDX) && !(tmp = get_child(tmp, flags))) {
			return -EIO;
		}
	} while (ntype == ND_INDX);

	if (ntype != ND_LEAF) {
		printk("hfs_bfind: invalid node type %02x in btree", HEAD(tmp)->ndType);
		hfs_brec_relse(&tmp);
		return -EIO;
	}

	if ((flags & BFIND_EXACT) && result) {
		hfs_brec_relse(&tmp);
		return -ENOENT;
	}

	if (flags & BFIND_LOCK) {
		brec_lock(tmp);
	}

	*brec = tmp;
	return 0;
}

/*======== Public Internal Functions ========*/

/*
 * This function releases a hfs_brec.
 * Currently the only way for a function outside of this source file
 * to obtain a new hfs_brec is through hfs_bfind()
 */
 void hfs_brec_relse(struct hfs_brec **brec) {
	struct hfs_brec *tmp;
#if defined(DEBUG_BRECS) || defined(DEBUG_ALL)
	extern int hfs_bcount;
#endif
	if (brec && (tmp = *brec) && (tmp->magic == HFS_BREC_MAGIC)) {
		if (tmp->bnode) {
			bnode_relse(tmp->bnode, tmp->lock_type);
		}
		hfs_brec_relse(&tmp->up);
		kfree_s(tmp, sizeof(struct hfs_brec));
		*brec = NULL;
#if defined(DEBUG_BRECS) || defined(DEBUG_ALL)
		--hfs_bcount;
#endif
	} else if (brec && (tmp = *brec)) {
		printk("hfs_brec_relse: corrupted hfs_brec.\n");
	}
}

/*
 * This function frees an hfs_btree structure obtained from hfs_btree_init().
 */
void hfs_btree_free(struct hfs_btree *bt)
{
	int lcv;
	struct hfs_bnode *bn, *tmp;

	if (bt && (bt->magic == HFS_BTREE_MAGIC)) {
		/* remove all cached nodes for this tree */
		for (lcv=0; lcv<CACHELEN; ++lcv) {
			bn = bnode_cache[lcv];
			while (bn != NULL) {
				tmp = bn->next;
				if (bn->tree == bt) {
#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
					printk("deleting node %d from tree %d with count %d, slot %d\n", bn->node, bn->fid, bn->count, lcv);
#endif
					bnode_delete(bn);
				}
				bn = tmp;
			}
		}
		bn = bt->head;
		while (bn != NULL) {
			tmp = bn->next;
#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
			printk("deleting node %d from tree %d with count %d\n", bn->node, bn->fid, bn->count);
#endif
			bnode_delete(bn);
			bn = tmp;
		}
#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
		printk("deleting node %d from tree %d with count %d\n", bt->root->node, bt->root->fid, bt->root->count);
#endif
		bnode_delete(bt->root);
		kfree_s(bt, sizeof(struct hfs_btree));
	} else if (bt) {
		printk("hfs_btree_free: corrupted hfs_btree.\n");
	} else {
		printk("hfs_btree_free: NULL argument.\n");
	}
}

/*
 * Given some vital information from the MDB (superblock), this function
 * initializes the fields of an hfs_btree structure from disk.
 *
 * It return 0 on sucess, and non-0 if there is a problem
 */
struct hfs_btree * hfs_btree_init(struct super_block *sb, unsigned int fid,
	const struct hfs_raw_extent *ext, HFS_LWORD tsize, HFS_LWORD csize)
{
	struct hfs_btree * bt;
	struct tree_head * th;
	struct hfs_bnode * tmp;
	unsigned int next;

	if (!sb || !ext || !(bt = kmalloc(sizeof(struct hfs_btree), GFP_KERNEL)))
		goto bail5;

	bt->magic = HFS_BTREE_MAGIC;
	bt->file.magic = HFS_FILE_MAGIC;
	bt->file.sb = sb;
	bt->file.fid = fid;
	bt->file.clumpsize = from_l(csize);
	bt->file.fork = HFS_FK_DATA;
	hfs_extent_in(&bt->file, ext);
	bt->dev = sb->s_dev;
	bt->tree_size = from_l(tsize) * HFS_SB(sb)->s_alloc_blksz;

	if (!(bt->head=kmalloc(sizeof(struct hfs_bnode), GFP_KERNEL)))
		goto bail4;
	memset(bt->head, 0, sizeof(struct hfs_bnode));
	bt->head->magic = HFS_BNODE_MAGIC;
	bt->head->tree = bt;
	bt->head->dev = bt->dev;
	bt->head->fid = fid;
	bt->head->node = 0;
	bt->head->sticky = 1;
	bt->head->prev = bt->head->next = NULL;
	if (!(bt->head->bh = bread(bt->dev, hfs_block_map(&bt->file, 0), HFS_SECTOR_SIZE)))
		goto bail3;
	th = (struct tree_head *)((char *)bt->head->bh->b_data + sizeof(struct node_head));

	/* read in the bitmap nodes (if any) */
	tmp = bt->head;
	while ((next = from_l(((struct node_head *)tmp->bh->b_data)->ndFLink))) {
		if (!(tmp->next=kmalloc(sizeof(struct hfs_bnode), GFP_KERNEL)))
			goto bail2;
		memset(tmp->next, 0, sizeof(struct hfs_bnode));
		tmp->next->prev = tmp;
		tmp = tmp->next;
		tmp->magic = HFS_BNODE_MAGIC;
		tmp->tree = bt;
		tmp->dev = bt->dev;
		tmp->fid = fid;
		tmp->node = next;
		tmp->sticky = 1;
		if (!(tmp->bh = bread(bt->dev, hfs_block_map(&bt->file, next), HFS_SECTOR_SIZE)))
			goto bail2;
	}

#if defined(DEBUG_HEADER) || defined(DEBUG_ALL)
	printk("tree_depth %d\n", (int)from_w(th->tree_depth));
	printk("root_node %d\n", (int)from_l(th->root_node));
	printk("leaf_recs %d\n", (int)from_l(th->leaf_recs));
	printk("first_leaf %d\n", (int)from_l(th->first_leaf));
	printk("last_leaf %d\n", (int)from_l(th->last_leaf));
	printk("node_size %d\n", (int)from_w(th->node_size));
	printk("key_len %d\n", (int)from_w(th->key_len));
	printk("total_nodes %d\n", (int)from_l(th->total_nodes));
	printk("free_nodes %d\n", (int)from_l(th->free_nodes));
#endif

	if (from_w(th->node_size) != HFS_SECTOR_SIZE) {
		printk("hfs_btree_init: node_size!=512 not supported\n");
		goto bail2;
	}

	if (fid == HFS_CAT_FID)
		bt->compf = cat_compare;
	else if (fid == HFS_EXT_FID)
		bt->compf = ext_compare;
	else
		goto bail2;
	bt->tree_depth = from_w(th->tree_depth);
	bt->root_node = from_l(th->root_node);
	bt->leaf_recs = from_l(th->leaf_recs);
	bt->first_leaf = from_l(th->first_leaf);
	bt->last_leaf = from_l(th->last_leaf);
	bt->key_len = from_w(th->key_len);
	bt->total_nodes = from_l(th->total_nodes);
	bt->free_nodes = from_l(th->free_nodes);

	if (!(bt->root=kmalloc(sizeof(struct hfs_bnode), GFP_KERNEL)))
		goto bail2;
	memset(bt->root, 0, sizeof(struct hfs_bnode));
	bt->root->magic = HFS_BNODE_MAGIC;
	bt->root->tree = bt;
	bt->root->dev = bt->dev;
	bt->root->fid = fid;
	bt->root->node = bt->root_node;
	bt->root->sticky = 1;
	bt->root->prev = bt->root->next = NULL;
	if (!(bt->root->bh = bread(bt->dev, hfs_block_map(&bt->file, bt->root_node), HFS_SECTOR_SIZE)))
		goto bail1;
	return bt;

bail1:
	kfree_s(bt->root, sizeof(struct hfs_bnode));
bail2:
	brelse(bt->head->bh);
	tmp = bt->head->next;
	while (tmp) {
		if (tmp->bh)
			brelse(tmp->bh);
		if (tmp->next) {
			tmp = tmp->next;
			kfree_s(tmp->prev, sizeof(struct hfs_bnode));
		} else {
			kfree_s(tmp, sizeof(struct hfs_bnode));
			tmp = NULL;
		}
	}
bail3:
	kfree_s(bt->head, sizeof(struct hfs_bnode));
bail4:
	kfree_s(bt, sizeof(struct hfs_btree));
bail5:
	return NULL;
}

/*
 * This is the wrapper for bfind() that allows access from outside this
 * source file.
 *
 * Its main purpose is to construct an allowable flags argument for bfind().
 */
int hfs_bfind(struct hfs_btree *tree, struct hfs_rec_key *target, struct hfs_brec **brec, int mode) {
	int flags;

	if (mode == HFS_BFIND_READ_EQ)
		flags = BFIND_RDONLY | BFIND_EXACT;
	else if (mode == HFS_BFIND_READ_LE)
		flags = BFIND_RDONLY;
	else if (mode == HFS_BFIND_WRITE)
		flags = BFIND_EXACT | BFIND_LOCK;
	else {
		printk("hfs_bfind: invalid mode %d\n", mode);
		return -EINVAL;
	}
	return bfind(tree, target, brec, flags);
}

/*
 * This function overwrites brec with its succesor.
 *
 * It returns 0 on success.
 * In case of failure brec is released if possible, and -errno is returned.
  */
int hfs_bsucc(struct hfs_brec **brec)
{
	struct hfs_brec *tmp;
	struct hfs_bnode *up;

	if (brec && (tmp = *brec) && (tmp->magic == HFS_BREC_MAGIC) && 
	    tmp->bnode && (tmp->bnode->magic == HFS_BNODE_MAGIC) &&
	    tmp->bnode->tree && (tmp->bnode->tree->magic == HFS_BTREE_MAGIC) &&
	    tmp->bnode->bh) {
		if (tmp->record == from_w(HEAD(tmp)->ndNRecs)) {
			/* get next node */
			/* XXX may eventually need to work with writes */
			up = tmp->bnode;
			tmp->bnode = read_node(tmp->bnode->tree, from_l(HEAD(tmp)->ndFLink), tmp->lock_type);
			bnode_relse(up, tmp->lock_type);
			if (!tmp->bnode) {
				printk("hfs_bsucc: unable to read node from dev 0x%04x\n",
					tmp->bnode->dev);
				hfs_brec_relse(brec);
				return -EIO;
			}
			tmp->record = 1;
			tmp->record_tbl = REC_TBL(tmp);
			tmp->key = KEY(tmp, 1);
		} else {
			tmp->record++;
			tmp->key = KEY(tmp, tmp->record);
		}
		tmp->data.generic = RECORD(tmp->key);
		return 0;
	}
	printk("hfs_bsucc: invalid/corrupt arguments.\n");
	return -EINVAL;
}
