/*
 *  linux/fs/hfs/dir.c
 *
 *  Copyright (C) 1995  Paul H. Hargrove
 *  This file may be distributed under the terms of the GPL
 *
 *  Based on the minix file system code, (C) 1991, 1992 by Linus Torvalds
 */

#include <linux/config.h>
#ifdef MODULE
#include <linux/module.h>
#endif

#include "hfs_fs.h"
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <asm/segment.h>
#include <linux/dirent.h>

#define NAME_OFFSET(de) ((int) ((de)->d_name - (char *) (de)))
#define ROUND_UP(x) (((x)+3) & ~3)
#define	mac_file_num(ARG)	(from_l((ARG->cdrType==HFS_CDR_FIL)?ARG->u.fil.FlNum:ARG->u.dir.DirID))
#define HFS_NAMEMAX	(3*31)

/*======== Private Internal Functions ========*/
#if 0
static __inline__ const int is_hex(char c) {
	return (((c>='0')&&(c<='9')) ||
		((c>='a')&&(c<='f')) ||
		((c>='A')&&(c<='F')));
}

static __inline__ int is_hdr(const char *name, int len) {
	return ((len > 1) && (name[0]=='%') && 
		((!is_hex(name[1])) || (len < 3) || (!is_hex(name[2]))));
}
#endif

static struct inode * dupdir(struct inode * dir, unsigned int type)
{
	struct inode *result = iget(dir->i_sb, HFS_IFID(dir->i_ino) | type);

	if (!result->i_mode) {
		memcpy(HFS_I(result), HFS_I(dir), sizeof(*HFS_I(dir)));
		result->i_mtime = dir->i_mtime;
		result->i_ctime = dir->i_ctime;
		result->i_atime = dir->i_atime;
		result->i_size = HFS_I(dir)->u.dir.files+2;
		result->i_nlink = 2;
		if (type != HFS_CAP_RDIR)
			result->i_size += HFS_I(dir)->u.dir.dirs;
		if (type == HFS_CAP_NDIR) {
			result->i_size += 2;
			result->i_nlink += (2 + HFS_I(dir)->u.dir.dirs);
		}
		result->i_mode = dir->i_mode;
		result->i_blocks = 0;
		result->i_op = dir->i_op;	/* not once r/w */
		result->i_uid = dir->i_uid;
		result->i_gid = dir->i_gid;
	}
	return result;
}

#ifdef LINUX_1_2
/* this goes away in 1.3.x, as Linus has adopted this approach gloably */
static int build_dirent(struct dirent *d, unsigned int ino, unsigned int len, char *name)
{
	if (len > 0) {
		put_fs_long(ino, &d->d_ino);
		memcpy_tofs(d->d_name, name, len);
		put_fs_byte(0, len+d->d_name);
		put_fs_word(len, &d->d_reclen);
		return ROUND_UP(NAME_OFFSET(d) + len + 1);
	} else {
		return 0;
	}
}
#endif /* LINUX_1_2 */

static int lookup_parent(struct inode *dir, struct inode **result)
{
	struct hfs_brec *entry;
	int error;

	error = hfs_cat_find(dir->i_sb, HFS_I(dir)->u.dir.parent, 0, NULL, &entry);
	if (!error) {
		*result = hfs_cap_iget(dir->i_sb, 
				from_l(entry->data.raw_inode->u.thd.ParID),
				entry->data.raw_inode->u.thd.CName[0],
				entry->data.raw_inode->u.thd.CName+1, HFS_ITYPE_NORM);
		hfs_brec_relse(&entry);
	}
	return error;
}

/*======== Filesystem Functions ========*/

static int hfs_dir_read(struct inode * inode, struct file * filp, char * buf, int count)
{
	return -EISDIR;
}

static int hfs_dbl_lookup(struct inode * dir, const char * name, int len,
	struct inode ** result)
{
	char tmp_name[HFS_NAMELEN];

	if (len==1 && name[0] == '.') {
		/* this little 'trick' skips the iget and iput */
		*result = dir;
		goto done;
	}
	if (len==2 && name[0] == '.' && name[1] == '.') {
		int error;

		error = lookup_parent(dir, result);
		if (error)
			return error;
	} else {
		if ((name[0] == '%') && ((len = HFS_SB(dir->i_sb)->s_nameout(tmp_name, len-1, name+1))>0)) {
			*result = hfs_dbl_iget(dir->i_sb, HFS_IFID(dir->i_ino), len, tmp_name, HFS_DBL_HDR);
		} 
		if ((!*result) && ((len = HFS_SB(dir->i_sb)->s_nameout(tmp_name, len, name))>0)) {
			*result = hfs_dbl_iget(dir->i_sb, HFS_IFID(dir->i_ino), len, tmp_name, HFS_DBL_NORM);
		} 
	}
	iput(dir);
	if (!*result)
		return -ENOENT;
done:
	return 0;
}

static int hfs_cap_lookup(struct inode * dir, const char * name, int len,
	struct inode ** result)
{
	int type = HFS_ITYPE(dir->i_ino);
	char tmp_name[HFS_NAMELEN];
	int tmp;

	if (len==1 && name[0] == '.') {
		/* this little 'trick' skips the iget and iput */
		*result = dir;
		goto done;
	} 
	if (len==2 && name[0] == '.' && name[1] == '.') {
		if (type != HFS_CAP_NDIR) {
			*result = dupdir(dir, HFS_CAP_NDIR);
		} else {
			int error;

			error = lookup_parent(dir, result);
			if (error)
				return error;
		}
	} else if ((type==HFS_CAP_NDIR) && len==9 && !strncmp(name, ".resource", 9)) {
		*result = dupdir(dir, HFS_CAP_RDIR);
	} else if ((type==HFS_CAP_NDIR) && len==11 && !strncmp(name, ".finderinfo", 11)) {
		*result = dupdir(dir, HFS_CAP_FDIR);
	} else if ((len=HFS_SB(dir->i_sb)->s_nameout(tmp_name, len, name)) > 0) {
		switch (type) {
		case HFS_CAP_RDIR:
			tmp = HFS_CAP_RSRC;
			break;

		case HFS_CAP_FDIR:
			tmp = HFS_CAP_FNDR;
			break;

		default:
			tmp = HFS_CAP_NORM;
		}
		*result = hfs_cap_iget(dir->i_sb, HFS_IFID(dir->i_ino), len, tmp_name, tmp);
	}
	iput(dir);
	if (!*result)
		return -ENOENT;
done:
	return 0;
}

#ifndef LINUX_1_2
static int hfs_dbl_readdir(struct inode * dir, struct file * filp,
	void * dirent, filldir_t filldir)
#else /* LINUX_1_2 */
static int hfs_dbl_readdir(struct inode * dir, struct file * filp,
	struct dirent * dirent, int count)
#endif /* LINUX_1_2 */
/* XXX we should probably see about hiding the meta-data of covered dirs */
{
	int res, old_pos, pos, error, ino, version;
	unsigned int len, tmp;
	struct hfs_brec *entry;
	char fixed_name[HFS_NAMEMAX+1];

	ino = dir->i_ino;

	if (filp->f_pos >= dir->i_size)
		return 0;
	old_pos = filp->f_pos++;
	do {
		pos = old_pos;
		version = dir->i_version;
		if (!pos) {
			/* entry for '.' */
#ifndef LINUX_1_2
			res = filldir(dirent, ".", 1, 0, ino);
#else /* LINUX_1_2 */
			res = build_dirent(dirent, ino, 1, ".");
#endif /* LINUX_1_2 */
		} else if (pos==1) {
			/* entry for '..' */
#ifndef LINUX_1_2
			res = filldir(dirent, "..", 2, 1, HFS_I(dir)->u.dir.parent);
#else /* LINUX_1_2 */
			res = build_dirent(dirent, HFS_I(dir)->u.dir.parent, 2, "..");
#endif /* LINUX_1_2 */
		} else {
			error = hfs_cat_find(dir->i_sb, HFS_IFID(ino), 0 ,NULL, &entry);
			if (!error)
				error = hfs_cat_next(&entry, pos>>1, 0);
			if (error)
				return error;
			if ((entry->data.raw_inode->cdrType!=HFS_CDR_FIL)&&
	    		(entry->data.raw_inode->cdrType!=HFS_CDR_DIR)) {
				printk("hfs_readdir: catalog entry is neither file nor directory\n");
				hfs_brec_relse(&entry);
				return -ENOENT;
			}
			tmp = mac_file_num(entry->data.raw_inode);
			if (pos & 1) {
				/* header */
				fixed_name[0] = '%';
				len = 1 + HFS_SB(dir->i_sb)->s_namein(fixed_name+1, entry->key->u.ckr.CName);
			} else {
				/* normal */
				len = HFS_SB(dir->i_sb)->s_namein(fixed_name, entry->key->u.ckr.CName);
			}
			hfs_brec_relse(&entry);
#ifndef LINUX_1_2
			res = filldir(dirent, fixed_name, len, old_pos, tmp);
#else /* LINUX_1_2 */
			res = build_dirent(dirent, tmp, len, fixed_name);
#endif /* LINUX_1_2 */
		}
	} while (version != dir->i_version);
	return res;
}

#ifndef LINUX_1_2
static int hfs_cap_readdir(struct inode * dir, struct file * filp,
	void * dirent, filldir_t filldir)
#else /* LINUX_1_2 */
static int hfs_cap_readdir(struct inode * dir, struct file * filp,
	struct dirent * dirent, int count)
#endif /* LINUX_1_2 */
{
	int res, old_pos, pos, error, skip, ino, version;
	unsigned int len, tmp;
	struct hfs_brec *entry;
	char fixed_name[HFS_NAMEMAX];

	ino = dir->i_ino;

	if (filp->f_pos >= dir->i_size)
		return 0;
	old_pos = filp->f_pos++;
	do {
		pos = old_pos;
		version = dir->i_version;
		if (!pos) {
			/* entry for '.' */
#ifndef LINUX_1_2
			res = filldir(dirent, ".", 1, 0, ino);
#else /* LINUX_1_2 */
			res = build_dirent(dirent, ino, 1, ".");
#endif /* LINUX_1_2 */
		} else if (!(--pos)) {
			/* entry for '..' */
#ifndef LINUX_1_2
			res = filldir(dirent, "..", 2, 1, HFS_I(dir)->u.dir.parent);
#else /* LINUX_1_2 */
			res = build_dirent(dirent, HFS_I(dir)->u.dir.parent, 2, "..");
#endif /* LINUX_1_2 */
		} else if ((HFS_ITYPE(ino)==HFS_CAP_NDIR) && !(--pos)) {
			/* entry for '.resource' */
#ifndef LINUX_1_2
			res = filldir(dirent, ".resource", 9, old_pos, HFS_IFID(ino)|HFS_CAP_RDIR);
#else /* LINUX_1_2 */
			res = build_dirent(dirent, HFS_IFID(ino)|HFS_CAP_RDIR, 9, ".resource");
#endif /* LINUX_1_2 */
		} else if ((HFS_ITYPE(ino)==HFS_CAP_NDIR) && !(--pos)) {
			/* entry for '.finderinfo' */
#ifndef LINUX_1_2
			res = filldir(dirent, ".finderinfo", 11, old_pos, HFS_IFID(ino)|HFS_CAP_FDIR);
#else /* LINUX_1_2 */
			res = build_dirent(dirent, HFS_IFID(ino)|HFS_CAP_FDIR, 11, ".finderinfo");
#endif /* LINUX_1_2 */
		} else {
			if (HFS_ITYPE(ino)==HFS_CAP_RDIR) /* no resource forks for directories */
				skip = HFS_SKIP_DIR;
			else
				skip = 0;
			error = hfs_cat_find(dir->i_sb, HFS_IFID(ino), 0 ,NULL, &entry);
			if (!error)
				error = hfs_cat_next(&entry, pos, skip);
			if (error)
				return error;
			if ((entry->data.raw_inode->cdrType!=HFS_CDR_FIL)&&
	    		(entry->data.raw_inode->cdrType!=HFS_CDR_DIR)) {
				printk("hfs_readdir: catalog entry is neither file nor directory\n");
				hfs_brec_relse(&entry);
				return -ENOENT;
			}
			tmp = mac_file_num(entry->data.raw_inode);
			len = HFS_SB(dir->i_sb)->s_namein(fixed_name, entry->key->u.ckr.CName);
			hfs_brec_relse(&entry);
#ifndef LINUX_1_2
			res = filldir(dirent, fixed_name, len, old_pos, tmp);
#else /* LINUX_1_2 */
			res = build_dirent(dirent, tmp, len, fixed_name);
#endif /* LINUX_1_2 */
		}
	} while (version != dir->i_version);
	return res;
}

/*======== Filesystem Function Tables ========*/

static struct file_operations hfs_cap_dir_operations = {
	NULL,			/* lseek - default */
	hfs_dir_read,		/* read - invalid */
	NULL,			/* write - bad */
	hfs_cap_readdir,	/* readdir */
	NULL,			/* select - default */
	NULL,			/* ioctl - default */
	NULL,			/* no special open code */
	NULL,			/* no special release code */
	NULL			/* fsync */
};

struct inode_operations hfs_cap_dir_inode_operations = {
	&hfs_cap_dir_operations,/* default directory file-ops */
	NULL,      	        /* create */
	hfs_cap_lookup,		/* lookup */
	NULL, 		        /* link */
	NULL,       		/* unlink */
	NULL,       		/* symlink */
	NULL,       		/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL,			/* bmap */
	NULL,			/* truncate */
	NULL,			/* permission */
	NULL			/* smap */
};

static struct file_operations hfs_dbl_dir_operations = {
	NULL,			/* lseek - default */
	hfs_dir_read,		/* read - invalid */
	NULL,			/* write - bad */
	hfs_dbl_readdir,	/* readdir */
	NULL,			/* select - default */
	NULL,			/* ioctl - default */
	NULL,			/* no special open code */
	NULL,			/* no special release code */
	NULL			/* fsync */
};

struct inode_operations hfs_dbl_dir_inode_operations = {
	&hfs_dbl_dir_operations,/* default directory file-ops */
	NULL,      	        /* create */
	hfs_dbl_lookup,		/* lookup */
	NULL, 		        /* link */
	NULL,       		/* unlink */
	NULL,       		/* symlink */
	NULL,       		/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL,			/* bmap */
	NULL,			/* truncate */
	NULL,			/* permission */
	NULL			/* smap */
};
