/*
 *  linux/fs/hfs/extent.c
 *
 *  Copyright (C) 1995  Paul H. Hargrove
 *  This file may be distributed under the terms of the GPL
 *
 */

#include <linux/config.h>
#ifdef MODULE
#include <linux/module.h>
#endif

#include "hfs_fs.h"
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/fs.h>


/*======== Private Internal Functions ========*/

#if defined(DEBUG_EXTENTS) || defined(DEBUG_ALL)
static void dump_ext(const struct hfs_extent *e) {
	printk("(%d:%d) (%d:%d) (%d:%d)\n", e->start, e->block1,
		e->start + e->length1, e->block2,
		e->start + e->length1 + e->length2, e->block3);
}
#endif

struct raw {
	HFS_WORD	block1		DONT_ALIGN;
	HFS_WORD	length1		DONT_ALIGN;
	HFS_WORD	block2		DONT_ALIGN;
	HFS_WORD	length2		DONT_ALIGN;
	HFS_WORD	block3		DONT_ALIGN;
	HFS_WORD	length3		DONT_ALIGN;
};

/* all callers check to,from != NULL */
static void read_extent(struct hfs_extent *to, const struct hfs_raw_extent *from,
	unsigned short start)
{
	const struct raw *tmp = (const struct raw *)from;

	to->start = start;
	to->block1 = from_w(tmp->block1);
	to->length1 = from_w(tmp->length1);
	to->block2 = from_w(tmp->block2);
	to->length2 = from_w(tmp->length2);
	to->block3 = from_w(tmp->block3);
	to->length3 = from_w(tmp->length3);
}

/* all callers check to,from != NULL */
static void write_extent(struct hfs_raw_extent *to, const struct hfs_extent *from)
{
	struct raw *tmp = (struct raw *)to;

	tmp->block1 = to_w(from->block1);
	tmp->length1 = to_w(from->length1);
	tmp->block2 = to_w(from->block2);
	tmp->length2 = to_w(from->length2);
	tmp->block3 = to_w(from->block3);
	tmp->length3 = to_w(from->length3);
}

/*
 * Take an extent record and allocation block offset into the file,
 * and try to return the number of the corresponding allocation block
 * on disk. If the block is beyond the current extent record return -1.
 */
/* all callers check extent!=NULL */
static int decode_extent(const struct hfs_extent * extent, int block)
{
	if (block < extent->start)
		return -1;
	block -= extent->start;
	if (block < extent->length1)
		return block + extent->block1;
	block -= extent->length1;
	if (block < extent->length2)
		return block + extent->block2;
	block -= extent->length2;
	if (block < extent->length3)
		return block + extent->block3;
	return -1;
}

/*
 * Given a FileID and an allocation block number, find the extent record
 * containing that block, return zero on success, nonzero on failure
 */
/* all callers check tr,fil != NULL */
static __inline__ int find_ext(struct hfs_btree *tr, struct hfs_file *fil, int alloc_block)
{
	struct hfs_rec_key target;
	struct hfs_brec	*entry;
	int retval = 1;

	if (fil->fid != HFS_EXT_FID) {
		/*target.KeyLen = XXX*/
		target.u.xkr.FkType = fil->fork;
		target.u.xkr.FNum = to_l(fil->fid);
		target.u.xkr.FABN = to_w(alloc_block);

		if (!hfs_bfind(tr, &target, &entry, HFS_BFIND_READ_LE)) {
			if ((entry->key->u.xkr.FNum == target.u.xkr.FNum) &&
			    (entry->key->u.xkr.FkType == fil->fork)) {
				read_extent(&fil->cache, entry->data.raw_extent, from_w(entry->key->u.xkr.FABN));
				retval = 0;
			} 
			hfs_brec_relse(&entry);
		}
	} else {
		printk("hfs_block_map: bad extents tree: infinite recursion\n");
	}
	return retval;
}

/*======== Public Internal Functions ========*/

int hfs_block_map(struct hfs_file *fil, int block) 
{
	int ablksz, offset, tmp;

	if (!fil || !fil->sb)
		return 0;
#if defined(DEBUG_EXTENTS) || defined(DEBUG_ALL)
	printk("block_map: ablock %d of file %d, fork %d", block, fil->fid, fil->fork);
#endif

	if (block < 0) {
		printk("hfs_block_map: block < 0\n");
		return 0;
	}
	ablksz = HFS_SB(fil->sb)->s_alloc_blksz;
	offset = HFS_SB(fil->sb)->s_fs_start + (block % ablksz);
	block /= ablksz;
	if (block > HFS_USHRT_MAX) {
		printk("hfs_block_map: block > big\n");
		return 0;
	}
#if defined(DEBUG_EXTENTS) || defined(DEBUG_ALL)
	printk("(lblock %d offset %d)\n", block, offset);
	printk("trying cache: ");
	dump_ext(&fil->cache);
#endif
	if ((tmp = decode_extent(&fil->cache, block)) >= 0)
		return tmp * ablksz + offset;
#if defined(DEBUG_EXTENTS) || defined(DEBUG_ALL)
	printk("trying first: ");
	dump_ext(&fil->first);
#endif
	if ((tmp = decode_extent(&fil->first, block)) >= 0) {
		memcpy(&fil->cache, &fil->first, sizeof(struct hfs_extent));
		return tmp * ablksz + offset;
	}
	if (find_ext(HFS_SB(fil->sb)->s_ext_tree, fil, block))
		return 0;
#if defined(DEBUG_EXTENTS) || defined(DEBUG_ALL)
	printk("trying new: ");
	dump_ext(&fil->first);
#endif
	if ((tmp = decode_extent(&fil->cache, block)) >= 0)
		return tmp*ablksz + offset;
	return 0;
}

void hfs_extent_out(const struct hfs_file *fil, struct hfs_raw_extent *ext)
{
	if (fil && ext) {
		write_extent(ext, &fil->first);
#if defined(DEBUG_EXTENTS) || defined(DEBUG_ALL)
		printk("extent_out: ");
		dump_ext(&fil->first);
#endif
	}
}

void hfs_extent_in(struct hfs_file *fil, const struct hfs_raw_extent *ext)
{
	if (fil && ext) {
		read_extent(&fil->first, ext, 0);
		memcpy(&fil->cache, &fil->first, sizeof(struct hfs_extent));
#if defined(DEBUG_EXTENTS) || defined(DEBUG_ALL)
		printk("extent_in: ");
		dump_ext(&fil->first);
#endif
	}
}
