/*
 *  linux/fs/hfs/file.c
 *
 *  Copyright (C) 1995  Paul H. Hargrove
 *  This file may be distributed under the terms of the GPL
 *
 *  Based on the minix file system code, (C) 1991, 1992 by Linus Torvalds
 */

#include <linux/config.h>
#ifdef MODULE
#include <linux/module.h>
#endif

#include "hfs_fs.h"
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/stat.h>
#include <linux/sched.h>
#include <linux/locks.h>
#include <asm/segment.h>

#define NBUF 32

/*======== Private Internal Functions ========*/

static __inline__ struct buffer_head * hfs_getblk(struct hfs_file *fil, int block, int create)
{
	if (create) {
		printk("hfs_getblk: create!=0 not yet available\n");
		return NULL;
	}
	block = hfs_block_map(fil, block);
	if (!block) {
		printk("hfs_getblk: hfs_block_map failed\n");
		return NULL;
	}
	return getblk(fil->sb->s_dev, block, HFS_SECTOR_SIZE);
}

static int do_read(struct hfs_file * fil, struct file * filp, int offset,
	 int count, unsigned int size, char * buf)
/*
 * This is just the "guts" of Linus's minix_file_read with
 * "minix"->"hfs" and "BLOCK_SIZE"->"HFS_SECTOR_SIZE" everywhere,
 * and a few other minor changes.
 */
{
	int read, chars;
	int block, blocks;
	int bhrequest, uptodate;
	struct buffer_head ** bhb, ** bhe;
	struct buffer_head * bhreq[NBUF];
	struct buffer_head * buflist[NBUF];

	read = 0;
	block = offset >> HFS_SECTOR_SIZE_BITS;
	offset &= HFS_SECTOR_SIZE-1;
	size = (size + (HFS_SECTOR_SIZE-1)) >> HFS_SECTOR_SIZE_BITS;
	blocks = (count + offset + HFS_SECTOR_SIZE - 1) >> HFS_SECTOR_SIZE_BITS;
	bhb = bhe = buflist;
	if (filp->f_reada) {
	        if(blocks < read_ahead[MAJOR(fil->sb->s_dev)] / (HFS_SECTOR_SIZE >> 9))
		  blocks = read_ahead[MAJOR(fil->sb->s_dev)] / (HFS_SECTOR_SIZE >> 9);
		if (block + blocks > size)
			blocks = size - block;
	}

	/* We do this in a two stage process.  We first try and request
	   as many blocks as we can, then we wait for the first one to
	   complete, and then we try and wrap up as many as are actually
	   done.  This routine is rather generic, in that it can be used
	   in a filesystem by substituting the appropriate function in
	   for getblk.

	   This routine is optimized to make maximum use of the various
	   buffers and caches. */

	do {
		bhrequest = 0;
		uptodate = 1;
		while (blocks) {
			--blocks;
			*bhb = hfs_getblk(fil, block++, 0);
			if (*bhb && !(*bhb)->b_uptodate) {
				uptodate = 0;
				bhreq[bhrequest++] = *bhb;
			}

			if (++bhb == &buflist[NBUF])
				bhb = buflist;

			/* If the block we have on hand is uptodate, go ahead
			   and complete processing. */
			if (uptodate)
				break;
			if (bhb == bhe)
				break;
		}

		/* Now request them all */
		if (bhrequest)
			ll_rw_block(READ, bhrequest, bhreq);

		do { /* Finish off all I/O that has actually completed */
			if (*bhe) {
				wait_on_buffer(*bhe);
				if (!(*bhe)->b_uptodate) {	/* read error? */
				        brelse(*bhe);
					if (++bhe == &buflist[NBUF])
					  bhe = buflist;
					count = 0;
					break;
				}
			}
			if (count < HFS_SECTOR_SIZE - offset)
				chars = count;
			else
				chars = HFS_SECTOR_SIZE - offset;
			filp->f_pos += chars;
			count -= chars;
			read += chars;
			if (*bhe) {
				memcpy_tofs(buf, offset+(*bhe)->b_data, chars);
				brelse(*bhe);
				buf += chars;
			} else {
				while (chars-->0)
					put_fs_byte(0,buf++);
			}
			offset = 0;
			if (++bhe == &buflist[NBUF])
				bhe = buflist;
		} while (count > 0 && bhe != bhb && (!*bhe || !(*bhe)->b_lock));
	} while (count > 0);

/* Release the read-ahead blocks */
	while (bhe != bhb) {
		brelse(*bhe);
		if (++bhe == &buflist[NBUF])
			bhe = buflist;
	};
	if (!read)
		return -EIO;
	filp->f_reada = 1;
	if (fil->inode && !IS_RDONLY(fil->inode))
		fil->inode->i_atime = CURRENT_TIME;
	return read;
}

/*======== File System Operations: Generic versions ========*/

static int hfs_bmap(struct inode * inode, int block)
{
	return hfs_block_map(&HFS_I(inode)->u.file, block);
}

static int hfs_file_read(struct inode * inode, struct file * filp, char * buf,
	 int count)
{
	int left, offset;
	unsigned int size;

	if (!inode) {
		printk("hfs_file_read: inode = NULL\n");
		return -EINVAL;
	}
	if (!S_ISREG(inode->i_mode)) {
		printk("hfs_file_read: mode = %07o\n",inode->i_mode);
		return -EINVAL;
	}
	offset = filp->f_pos;
	size = inode->i_size;
	if (offset > size)
		left = 0;
	else
		left = size - offset;
	if (left > count)
		left = count;
	if (left <= 0)
		return 0;
	return do_read(&HFS_I(inode)->u.file, filp, offset, left, size, buf);
}

/*======== File System Operations: AD versions ========*/

static int hfs_dbl_hdr_read(struct inode * inode, struct file * filp, char * buf, int count)
{
	unsigned int size, mem_count;
	int left, offset;

	if (!inode) {
		printk("hfs_file_read: inode = NULL\n");
		return -EINVAL;
	}
	if (!S_ISREG(inode->i_mode)) {
		printk("hfs_file_read: mode = %07o\n",inode->i_mode);
		return -EINVAL;
	}
	offset = filp->f_pos;
	size = inode->i_size;
	if (offset > size)
		left = 0;
	else
		left = size - offset;
	if (left > count)
		left = count;
	if (left <= 0)
		return 0;

	/* do the header if applicable */
	mem_count = 0;
	if (offset < HFS_HDR_LEN) {
		mem_count = HFS_HDR_LEN - offset;
		if (left < mem_count)
			mem_count = left;
		memcpy_tofs(buf, HFS_I(inode)->u.dbl_hdr.meta + offset, mem_count);
		filp->f_pos += mem_count;
	}
	if ((left -= mem_count) <= 0)
		return mem_count;
	offset -= HFS_HDR_LEN;
	size -= HFS_HDR_LEN;
	buf += mem_count;

	return mem_count + do_read(&HFS_I(inode)->u.dbl_hdr.rsrc, filp, offset, left, size, buf);
}

/*======== File System Operations Tables ========*/

static struct file_operations hfs_dbl_hdr_operations = {
	NULL,			/* lseek - default */
	hfs_dbl_hdr_read,	/* read */
	NULL,			/* write */
	NULL,			/* readdir - bad */
	NULL,			/* select - default */
	NULL,			/* ioctl - default */
	NULL, /*XXX*/		/* mmap - not yet */
	NULL,			/* no special open is needed */
	NULL,			/* release */
	NULL			/* fsync */
};

struct inode_operations hfs_dbl_hdr_inode_operations = {
	&hfs_dbl_hdr_operations,/* default file operations */
	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL, 			/* bmap - not avaiable */
	NULL,	       		/* truncate */
	NULL,			/* permission */
	NULL			/* smap */
};

static struct file_operations hfs_file_operations = {
	NULL,			/* lseek - default */
	hfs_file_read,		/* read */
	NULL,			/* write */
	NULL,			/* readdir - bad */
	NULL,			/* select - default */
	NULL,			/* ioctl - default */
	generic_mmap,		/* mmap */
	NULL,			/* no special open is needed */
	NULL,			/* release */
	NULL			/* fsync */
};

struct inode_operations hfs_file_inode_operations = {
	&hfs_file_operations,	/* default file operations */
	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	hfs_bmap,		/* bmap */
	NULL,	       		/* truncate */
	NULL,			/* permission */
	NULL			/* smap */
};
