/* 
 *  linux/include/linux/hfs_fs.h
 *
 *  Copyright (C) 1995 by Paul H. Hargrove
 *  This file may be distributed under the terms of the GPL
 */

#ifndef _HFS_FS_H
#define _HFS_FS_H

#include <linux/fs.h>
#include <asm/types.h>
#include <asm/byteorder.h>

#include "hfs_fs_i.h"
#include "hfs_fs_sb.h"

/* Macros to access hfs_inode_info and hfs_sb_info fields */
#define HFS_I(inode)		((struct hfs_inode_info *)(&inode->u))
#define HFS_SB(super)		((struct hfs_sb_info *)(&super->u))

/* offsets to various blocks */
#define HFS_PMAP_BLK		1 /* First block of partition map */
#define HFS_MDB_BLK		2 /* Block (w/i partition) of MDB */

/* magic numbers for various disk blocks */
#define HFS_DRVR_DESC_MAGIC	0x4552 /* driver descriptor map */
#define HFS_OLD_PMAP_MAGIC	0x5453 /* old-type partition map */
#define HFS_NEW_PMAP_MAGIC	0x504D /* new-type partition map */
#define HFS_SUPER_MAGIC		0x4244 /* HFS MDB (super block) */
#define HFS_MFS_SUPER_MAGIC	0xD2D7 /* MFS MDB (super block) */

/* magic numbers for CAP structures */
#define HFS_CAP_MAGIC		0xFF
#define HFS_CAP_MAGIC1		0xDA
#define HFS_CAP_VERSION		0x10

/* magic numbers for Apple Double header files */
#define HFS_DBL_MAGIC		0x00051607
#define HFS_DBL_VERSION		0x00020000

/* magic numbers for various internal structures */
#define HFS_FILE_MAGIC		0x4801
#define HFS_BREC_MAGIC		0x4802
#define HFS_BTREE_MAGIC		0x4803
#define HFS_BNODE_MAGIC		0x4804

/* various FIXED size parameters */
#define HFS_SECTOR_SIZE		512  /* size of an HFS sector */
#define HFS_SECTOR_SIZE_BITS	9    /* log_2(HFS_SECTOR_SIZE) */
#define HFS_NAMELEN		31   /* maximum length of an HFS filename */

/* The space used for the Apple Double or Apple Single headers */
#define HFS_HDR_LEN		1024

/* 2**16 - 1 */
#define	HFS_USHRT_MAX	65535

/* File ID numbers for the extents and catalog b-tree files */
#define	HFS_EXT_FID	3
#define	HFS_CAT_FID	4

/* Macros to extract file ID and file "type" from the Linux inode number */
#define HFS_IFID(X)	((X) & 0x3FFFFFFF)
#define HFS_ITYPE(X)	((X) & 0xC0000000)

/* How many bits used for the ITYPE */
#define HFS_ITYPE_BITS	2
#define HFS_ITYPE_SHIFT	(32-HFS_ITYPE_BITS)

/* generic ITYPEs */
#define HFS_ITYPE_NORM	0x00000000	/* "normal" directory or file */

/* ITYPEs for CAP */
#define HFS_CAP_DATA	HFS_ITYPE_NORM	/* data fork of file */
#define HFS_CAP_RSRC	0x80000000	/* resource fork of file */
#define HFS_CAP_FNDR	0x40000000	/* finder info for file or dir */
#define HFS_CAP_NDIR	HFS_ITYPE_NORM	/* normal directory */
#define HFS_CAP_RDIR	0x80000000	/* .resource directory */
#define HFS_CAP_FDIR	0xC0000000	/* .finderinfo directory */
#define HFS_CAP_NORM	HFS_ITYPE_NORM	/* data fork or normal directory */

/* ITYPEs for Apple Double */
#define HFS_DBL_DATA	HFS_ITYPE_NORM	/* data fork of file */
#define HFS_DBL_HDR	0x80000000	/* AD header of file or dir */
#define HFS_DBL_DIR	HFS_ITYPE_NORM	/* directory */
#define HFS_DBL_NORM	HFS_ITYPE_NORM	/* data fork or directory */

/* ITYPEs for Apple Single */
#define HFS_SNGL_SNGL	HFS_ITYPE_NORM	/* AppleSingle file */
#define HFS_SNGL_DINF	0x80000000	/* %DirInfo for directory */
#define HFS_SNGL_DIR	HFS_ITYPE_NORM	/* directory */
#define HFS_SNGL_NORM	HFS_ITYPE_NORM	/* data fork or directory */

/* Flags to hfs_cat_next */
#define HFS_SKIP_THD	0x01	/* exclude thread records */
#define HFS_SKIP_DIR	0x02	/* exclude directory records */

/* Mode for hfs_bfind */
enum hfs_bfind_mode {
	HFS_BFIND_READ_EQ,	/* find the exact entry */
	HFS_BFIND_READ_LE,	/* find largest entry <= key */
	HFS_BFIND_WRITE		/* find the exact entry, locked for writing */
};

/* Typedefs for the 68000's data types */
typedef	__u8		HFS_BYTE;
typedef __u16		HFS_WORD;
typedef __u32		HFS_LWORD;
typedef unsigned char	HFS_FNAME[HFS_NAMELEN+1];

/* Macro to prevent GCC from playing with alignment of HFS structures */
#define DONT_ALIGN	__attribute__ ((packed))

/* Byte-order conversions for 16-bit "word" */
#define from_w(X)	(ntohs(X))
#define to_w(X)		((HFS_WORD)htons(X))
#define c_to_w(X)	((HFS_WORD)__constant_htons(X))

/* Byte-order conversions for 32-bit "long word" */
#define from_l(X)	(ntohl(X))
#define to_l(X)		((HFS_LWORD)htonl(X))
#define c_to_l(X)	((HFS_LWORD)__constant_htonl(X))

/* 
 * host byte-order independent type promotions
 */
extern __inline__ HFS_LWORD word_to_long(HFS_WORD w) {
	union {
		HFS_WORD	words[2]	DONT_ALIGN;
		HFS_LWORD	lword		DONT_ALIGN;
	} x;
	x.lword = 0;
	x.words[1] = w;
	return x.lword;
}

extern __inline__ HFS_LWORD byte_to_long(HFS_BYTE b) {
	union {
		HFS_BYTE	bytes[4]	DONT_ALIGN;
		HFS_LWORD	lword		DONT_ALIGN;
	} x;
	x.lword = 0;
	x.bytes[3] = b;
	return x.lword;
}

extern __inline__ HFS_WORD byte_to_word(HFS_BYTE b) {
	union {
		HFS_BYTE	bytes[2]	DONT_ALIGN;
		HFS_WORD	word		DONT_ALIGN;
	} x;
	x.word = 0;
	x.bytes[1] = b;
	return x.word;
}

/*======== HFS structures as they appear on the disk ========*/

/* An extent record on disk*/
struct hfs_raw_extent{
	char		opaque[12]	DONT_ALIGN;
};

/* A btree record key on disk */
struct hfs_rec_key{
	/* The number of bytes in the key */
	HFS_BYTE	KeyLen				DONT_ALIGN;
	union {
		/* The key used in the catalog b-tree */
		struct hfs_cat_key {
			/* padding */
			HFS_BYTE	Resrv1		DONT_ALIGN;
			/* File ID of the directory containing the entry */
			HFS_LWORD	ParID		DONT_ALIGN;
			/* The filename of the entry */
			HFS_FNAME	CName		DONT_ALIGN;
		}	ckr				DONT_ALIGN;
		/* The key used in the extents b-tree */
		struct hfs_ext_key {
			/* fork type (HFS_FK_DATA or HFS_FK_RSRC) */
			HFS_BYTE	FkType		DONT_ALIGN;
			/* The File ID of the file */
			HFS_LWORD	FNum		DONT_ALIGN;
			/* offset (in allocation blocks) into the file */
			HFS_WORD	FABN		DONT_ALIGN;
		}	xkr				DONT_ALIGN;
	} u						DONT_ALIGN;
};

/* A catalog tree entry (equivalent to an inode) */
struct hfs_inode {
	/* The type of entry (see hfs_cdr_type, below) */
	HFS_BYTE	cdrType				DONT_ALIGN;
	/* padding */
	HFS_BYTE	cdrResrv2			DONT_ALIGN;
	union	{
		/* The entry for a file */
		struct {
			/* Flags such as read-only */
			HFS_BYTE	Flags		DONT_ALIGN;
			/* file version number, must be zero */
			HFS_BYTE	Typ		DONT_ALIGN;
			/* data used by FInder */
			HFS_BYTE	UsrWds[16]	DONT_ALIGN;
			/* The file ID */
			HFS_LWORD	FlNum		DONT_ALIGN;
			/* obsolete */
			HFS_WORD	StBlk		DONT_ALIGN;
			/* The logical EOF of the data fork*/
			HFS_LWORD	LgLen		DONT_ALIGN;
			/* The physical EOF of the data fork */
			HFS_LWORD	PyLen		DONT_ALIGN;
			/* obsolete */
			HFS_WORD	RStBlk		DONT_ALIGN;
			/* The logical EOF of the resource fork */
			HFS_LWORD	RLgLen		DONT_ALIGN;
			/* The physical EOF of the resource fork */
			HFS_LWORD	RPyLen		DONT_ALIGN;
			/* The creation date */
			HFS_LWORD	CrDat		DONT_ALIGN;
			/* The modified date */
			HFS_LWORD	MdDat		DONT_ALIGN;
			/* The last backup date */
			HFS_LWORD	BkDat		DONT_ALIGN;
			/* more date for the Finder */
			HFS_BYTE	FndrInfo[16]	DONT_ALIGN;
			/* number of bytes to allocate when extending */
			HFS_WORD	ClpSize		DONT_ALIGN;
			/* first extent record for the data fork */
			struct hfs_raw_extent ExtRec	DONT_ALIGN;
			/* first extent record for the resource fork */
			struct hfs_raw_extent RExtRec	DONT_ALIGN;
			/* reserved for future use by Apple */
			HFS_LWORD	Resrv		DONT_ALIGN;
		} fil					DONT_ALIGN;
		/* the entry for a directory */
		struct {
			/* flags */
			HFS_WORD	Flags		DONT_ALIGN;
			/* The number of files in the directory (Valence) */
			HFS_WORD	Val		DONT_ALIGN;
			/* The file ID of this directory */
			HFS_LWORD	DirID		DONT_ALIGN;
			/* The creation date */
			HFS_LWORD	CrDat		DONT_ALIGN;
			/* The modification date */
			HFS_LWORD	MdDat		DONT_ALIGN;
			/* The last backup date */
			HFS_LWORD	BkDat		DONT_ALIGN;
			/* data used by the Finder */
			HFS_BYTE	UsrInfo[16]	DONT_ALIGN;
			/* more data use by the Finder */
			HFS_BYTE	FndrInfo[16]	DONT_ALIGN;
			/* reserved for future use by Apple */
			HFS_BYTE	Resrv[16]	DONT_ALIGN;
		} dir					DONT_ALIGN;
		/* entry for a thread */
		/* Marks the beginning of the entries in a given directory */
		/* Serves the same purpose as a '..' directory entry */
		struct {
			/* reserved for future use by Apple */
			HFS_BYTE	Reserv[8]	DONT_ALIGN;
			/* file ID of the parent of this directory */
			HFS_LWORD	ParID		DONT_ALIGN;
			/* The name of this directory */
			HFS_FNAME	CName		DONT_ALIGN;
		} thd					DONT_ALIGN;
	} u						DONT_ALIGN;
};

/* legal values for hfs_inode.cdrType */
enum hfs_cdr_type {
	HFS_CDR_DIR = 1,
	HFS_CDR_FIL = 2,
	HFS_CDR_THD = 3,
	HFS_CDR_FTH = 4,
};

/* legal values for hfs_rec_key.u.xkr.FkType and hfs_file.fork */
enum hfs_fk_type {
	HFS_FK_DATA = 0x00,
	HFS_FK_RSRC = 0xFF,
};

/*======== Data structures kept in memory ========*/

/* 
 * This structure is used to refer to a b-tree record
 */
struct hfs_brec {
	/* A magic number */
	int                     magic;
	/* The node containing this record */
	struct hfs_bnode	*bnode;
	/* The index of this record within the node */
	int			record;
	/* The table of offsets to the records in this node */
	HFS_WORD		*record_tbl;
	/* A pointer to the key within the node */
	struct hfs_rec_key	*key;
	/* The type of lock held on the node (read-only/reserved/write) */
	int			lock_type;
	/* A doubly-linked list leading from the root to some leaf */
	struct hfs_brec		*up;
	struct hfs_brec		*down;
	/* The data associated with the key */
	union {
		/* To allow copying without knowing what type */
		void			*generic;
		/* In a leaf of the extents b-tree: */
		struct hfs_inode	*raw_inode;
		/* In a leaf of the catalog b-tree: */
		struct hfs_raw_extent	*raw_extent;
		/* In an internal node of either b-tree: */
		HFS_LWORD		*child;
	} data;
};

/*
 * A descriptor for a single entry within the header of an
 * Apple Double of Apple Single header file.
 * An array of these make up a table of contents for the file.
 */
struct hfs_dbl_descr {
	/* The Apple assigned ID for the type of entry */
	HFS_LWORD	id	DONT_ALIGN;
	/* The offset to reach the entry */
	HFS_LWORD	offset	DONT_ALIGN;
	/* The length of the entry */
	HFS_LWORD	length	DONT_ALIGN;
};

/* 
 * The header for an Apple Double header file.
 * There is space reserved for all the fields we can deal with.
 */
struct hfs_dbl_hdr {
	HFS_LWORD	magic		DONT_ALIGN;
	HFS_LWORD	version		DONT_ALIGN;
	HFS_BYTE	filler[16]	DONT_ALIGN;
	HFS_WORD	entries		DONT_ALIGN;
	struct hfs_dbl_descr
			descr[16]	DONT_ALIGN;
	HFS_LWORD	create_time	DONT_ALIGN;	/* \       */
	HFS_LWORD	modify_time	DONT_ALIGN;	/*  | id=8 */
	HFS_LWORD	backup_time	DONT_ALIGN;	/*  |      */
	HFS_LWORD	access_time	DONT_ALIGN;	/* /       */
	/* alignment doesn't matter past this point */
	HFS_BYTE	finderinfo[32];	/* id=9 */
	HFS_LWORD	fileinfo;	/* id=10 */
	HFS_FNAME	real_name;	/* id=3 */
	char		comment[200];	/* id=4 */
};

/* finder metadata for CAP */
struct hfs_cap_info {
	HFS_BYTE	fi_fndr[32]		DONT_ALIGN;
	HFS_WORD	fi_attr			DONT_ALIGN;
	HFS_BYTE	fi_magic1		DONT_ALIGN;
	HFS_BYTE	fi_version		DONT_ALIGN;
	HFS_BYTE	fi_magic		DONT_ALIGN;
	HFS_BYTE	fi_bitmap		DONT_ALIGN;
	HFS_BYTE	fi_shortfilename[12+1]	DONT_ALIGN;
	HFS_BYTE	fi_macfilename[32+1]	DONT_ALIGN;
	HFS_BYTE	fi_comln		DONT_ALIGN;
	HFS_BYTE	fi_comnt[200]		DONT_ALIGN;
};

/* string.c */
extern hfs_strcmp(const unsigned char *, const unsigned char *);

/* btree.c */
void hfs_brec_relse(struct hfs_brec **);
int hfs_bsucc(struct hfs_brec **);
int hfs_bfind(struct hfs_btree *, struct hfs_rec_key *, struct hfs_brec **, int);
struct hfs_btree *hfs_btree_init(struct super_block *, unsigned int, const struct hfs_raw_extent *, HFS_LWORD, HFS_LWORD);
void hfs_btree_free(struct hfs_btree *);

/* extent.c */
extern void hfs_extent_in(struct hfs_file *, const struct hfs_raw_extent *);
extern void hfs_extent_out(const struct hfs_file *, struct hfs_raw_extent *);
extern int hfs_block_map(struct hfs_file *, int);

/* catalog.c */
extern int hfs_cat_find(struct super_block *, unsigned int, int, const char *,
	struct hfs_brec **);
extern int hfs_cat_next(struct hfs_brec **, int, int);

/* inode.c */
extern struct inode * hfs_cap_iget(struct super_block *, unsigned int, int,
	const char *, int);
extern struct inode * hfs_dbl_iget(struct super_block *, unsigned int, int,
	const char *, int);
extern int hfs_notify_change(struct inode *, struct iattr *);
extern void hfs_cap_write_inode(struct inode *);
extern void hfs_dbl_write_inode(struct inode *);
#if 0
extern struct inode * hfs_sngl_iget(struct super_block *, unsigned int, int,
	const char *, int);
extern void hfs_sngl_write_inode(struct inode *);
#endif

/* super.c */
extern struct super_block *hfs_read_super(struct super_block *,void *,int);

/* file.c */
extern struct inode_operations hfs_file_inode_operations;
extern struct inode_operations hfs_dbl_hdr_inode_operations;
#if 0
extern struct inode_operations hfs_sngl_file_inode_operations;
#endif

/* dir.c */
extern struct inode_operations hfs_cap_dir_inode_operations;
extern struct inode_operations hfs_dbl_dir_inode_operations;
#if 0
extern struct inode_operations hfs_sngl_dir_inode_operations;
#endif

/* info.c */
extern struct inode_operations hfs_cap_info_inode_operations;

/* trans.c */
extern int hfs_cap2mac(char *, int, const char *);
extern int hfs_apl2mac(char *, int, const char *);
extern int hfs_triv2mac(char *, int, const char *);
extern int hfs_latin2mac(char *, int, const char *);
extern int hfs_mac2cap(char *, const unsigned char *);
extern int hfs_mac2latin(char *, const unsigned char *);
extern int hfs_mac2seven(char *, const unsigned char *);
extern int hfs_mac2eight(char *, const unsigned char *);
extern int hfs_mac2alpha(char *, const unsigned char *);
extern int hfs_mac2triv(char *, const unsigned char *);

#endif
