/* 
 * linux/include/linux/hfs_fs_i.h
 *
 * Copyright (C) 1995 by Paul H. Hargrove
 * This file may be distributed under the terms of the GPL
 *
 * This file defines the type (struct hfs_inode_info) and the two
 * subordinate types (struct hfs_extent) and (struct hfs_file).
 */

#ifndef _HFS_FS_I
#define _HFS_FS_I

/*
 * An extent record
 *
 * The offset to disk-block mapping for a given file is contained in a
 * series of these structures.  Each (struct hfs_extent) records upto
 * three runs of continuous disk-blocks.
 */
struct hfs_extent {
	/* The first allocation block within the file mapped in this record */
	unsigned short	start;
	/* The allocation block on disk which begins the first extent */
	unsigned short	block1;
	/* The number of allocation blocks in the first extent */
	unsigned short	length1;
	/* The allocation block on disk which begins the second extent */
	unsigned short	block2;
	/* The number of allocation blocks in the second extent */
	unsigned short	length2;
	/* The allocation block on disk which begins the third extent */
	unsigned short	block3;
	/* The number of allocation blocks in the third extent */
	unsigned short	length3;
};

/*
 * hfs "file" data 
 *
 * This structure hols the information necessary to locate a single
 * Linux file in an HFS filesystem.  Since a file has two forks, there
 * can be two different (struct hfs_file)s corresponding to the same
 * file on disk.  In this case, however, only the sb and fid are same
 * (and clumpsize most of the time).
 */
struct hfs_file {
	/* A magic number */
	int                magic;
	/* The filesystem containing this file */
	struct super_block *sb;
	/* The inode corresponding to this file. */
	/* In the cases of the catalog and extents b-trees, there is */
	/* no corresponding inode, and this field is NULL. */
	struct inode	   *inode;
	/* The first extent record for this file */
	struct hfs_extent  first;
	/* The most-recently accessed extent record for this file */
	struct hfs_extent  cache;
	/* The file id number for this file */
	unsigned int	   fid;
	/* The number of bytes to allocate when extending this file */
	unsigned short	   clumpsize;
	/* The fork (HFS_FK_DATA or HFS_FK_RSRC) of this file */
	unsigned char	   fork;
};

/*
 * hfs fs inode data in memory
 */
struct hfs_inode_info {
	/* The search key for the catalog tree entry for this file */
	/* This field (defined in hfs_fs.h) is in network byteorder */
  	unsigned char	key[38];
	/* The contents of the structure depends on whether this inode */
	/* is a regular file or a directory, and on how we are */
        /* representing forks. */
	union {
		/* data for a directory under all representations */
		struct {
			unsigned int		parent;
			unsigned short		dirs;
			unsigned short		files;
		} dir;
		/* data for a file under the CAP representation */
		/* and for the data fork under the Apple Double rep. */
		struct hfs_file file;
		/* data for the metadata file under the CAP rep. */
		struct {
			struct hfs_cap_info	*meta;
		} cap_info;
		/* data for the resource/metadata file under Apple Dbl */
		struct {
			struct hfs_dbl_hdr	*meta;
			struct hfs_file		rsrc;
		} dbl_hdr;
#if 0
		/* data for the file under the Apple Single representation */
		struct {
			struct hfs_dbl_hdr	*meta;
			struct hfs_file		rsrc;
			struct hfs_file		data;
		} sngl_file;
#endif
	} u;
};
#endif
