/*
 *  linux/fs/hfs/inode.c
 *
 *  Copyright (C) 1995  Paul H. Hargrove
 *  This file may be distributed under the terms of the GPL
 *
 *  Based on the minix file system code, (C) 1991, 1992 by Linus Torvalds
 */

#define FAKE_EXEC

#include <linux/config.h>
#ifdef MODULE
#include <linux/module.h>
#endif

#include "hfs_fs.h"
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/stat.h>
#include <linux/stddef.h>
#include <linux/malloc.h>
#include <linux/mm.h>

/*
 * There are three time systems.  All three are based on seconds since
 * a particular time/date.
 *	unix:	unsigned lil-endian since 00:00 GMT, Jan. 1, 1970
 *	mac:	unsigned big-endian since 00:00 GMT, Jan. 1, 1904
 *	header:	  SIGNED big-endian since 00:00 GMT, Jan. 1, 2000
 *
 */
#define	h_to_mtime(ARG)		to_l((signed long)from_l(ARG)+3029529600U)
#define	h_to_utime(ARG)		((signed long)from_l(ARG)+946684800U)
#define	m_to_htime(ARG)		((signed long)to_l(from_l(ARG)-3029529600U))
#define	m_to_utime(ARG)		(from_l(ARG)-2082844800U)
#define	u_to_htime(ARG)		((signed long)to_l((ARG)-946684800U))
#define	u_to_mtime(ARG)		to_l((ARG)+2082844800U)

#define	RAW	(entry->data.raw_inode)

static const struct hfs_dbl_descr hdr_template[3] = {
	/* descriptor for mac times: */
	{c_to_l(8), c_to_l(offsetof(struct hfs_dbl_hdr, create_time)), c_to_l(16)},
	/* descriptor for finder info: */
 	{c_to_l(9), c_to_l(offsetof(struct hfs_dbl_hdr, finderinfo)), c_to_l(32)},
	/* descriptor for file info flag word: */
 	{c_to_l(10), c_to_l(offsetof(struct hfs_dbl_hdr, fileinfo)), c_to_l(4)},
};

static __inline__ struct hfs_dbl_hdr *build_dbl_hdr(struct hfs_inode *raw) {
	struct hfs_dbl_hdr *res;
	HFS_LWORD mtime, ctime, btime;
	HFS_BYTE *info1, *info2;

	if ((res=(struct hfs_dbl_hdr *)kmalloc(HFS_HDR_LEN, GFP_KERNEL))) {
		memset(res, 0x00, HFS_HDR_LEN);
		res->magic = to_l(HFS_DBL_MAGIC);
		res->version = to_l(HFS_DBL_VERSION);
		memcpy(res->descr, &hdr_template, sizeof(hdr_template));
		if (raw->cdrType == HFS_CDR_FIL) {
			res->entries = to_w(4);
			res->descr[3].id = to_l(2);
			res->descr[3].offset = to_l(HFS_HDR_LEN);
			res->descr[3].length = raw->u.fil.RLgLen;
			res->fileinfo = byte_to_long(raw->u.fil.Flags);
			mtime = raw->u.fil.MdDat;
			ctime = raw->u.fil.CrDat;
			btime = raw->u.fil.BkDat;
			info1 = raw->u.fil.UsrWds;
			info2 = raw->u.fil.FndrInfo;
		} else {
			res->entries = to_w(3);
			res->fileinfo = word_to_long(raw->u.dir.Flags);
			mtime = raw->u.dir.MdDat;
			ctime = raw->u.dir.CrDat;
			btime = raw->u.dir.BkDat;
			info1 = raw->u.dir.UsrInfo;
			info2 = raw->u.dir.FndrInfo;
		}
		res->create_time = m_to_htime(ctime);
		res->modify_time = m_to_htime(mtime);
		res->access_time = res->modify_time;
		res->backup_time = m_to_htime(btime);
		memcpy(res->finderinfo, info1, 16);
		memcpy(res->finderinfo+16, info2, 16);
	}
	return res;
}

static __inline__ struct hfs_cap_info * build_cap_info(struct hfs_brec *entry)
{
	struct hfs_cap_info *res;
	HFS_BYTE *info1, *info2;

	if ((res = kmalloc(sizeof(struct hfs_cap_info),GFP_KERNEL))) {
		memset(res,0,sizeof(struct hfs_cap_info));
		if (RAW->cdrType==HFS_CDR_FIL) {
			info1 = RAW->u.fil.UsrWds;
			info2 = RAW->u.fil.FndrInfo;
		} else {
			info1 = RAW->u.dir.UsrInfo;
			info2 = RAW->u.dir.FndrInfo;
		}
		memcpy(res->fi_fndr, info1, 16);
		memcpy(res->fi_fndr+16, info2, 16);
		res->fi_magic1 = HFS_CAP_MAGIC1;
		res->fi_version = HFS_CAP_VERSION;
		res->fi_magic = HFS_CAP_MAGIC;
		res->fi_bitmap = 0x02;
		memcpy(res->fi_macfilename, entry->key->u.ckr.CName+1, entry->key->u.ckr.CName[0]);
	}
	return res;
}

static int count_entries(struct super_block *sb, unsigned int dir, int count,
	unsigned short *dirs, unsigned short *files)
{
	struct hfs_brec *entry;

	*dirs = *files = 0;
	if (hfs_cat_find(sb, dir, 0, NULL, &entry))
		return -1;

	while (count--) {
		if (hfs_cat_next(&entry, 1, 0))
			return -1;
		if (RAW->cdrType==HFS_CDR_FIL)
			(*files)++;
		else if (RAW->cdrType==HFS_CDR_DIR)
			(*dirs)++;
	};
	hfs_brec_relse(&entry);
	return 0;
}

static void read_file_inode(struct hfs_file *file, struct inode *inode,
	struct hfs_inode *raw, int fork, unsigned int ino)
{
	HFS_LWORD lsize, psize;
	struct hfs_raw_extent *ext;

	file->magic = HFS_FILE_MAGIC;
	file->sb = inode->i_sb;
	file->inode = inode;
	file->fid = ino;
	file->clumpsize = from_w(raw->u.fil.ClpSize);
	if ((file->fork = fork) == HFS_FK_DATA) {
		lsize = raw->u.fil.LgLen;
		psize = raw->u.fil.PyLen;
		ext = &raw->u.fil.ExtRec;
	} else {
		lsize = raw->u.fil.RLgLen;
		psize = raw->u.fil.RPyLen;
		ext = &raw->u.fil.RExtRec;
	}
	hfs_extent_in(file, ext);
#ifdef FAKE_EXEC
	inode->i_mode = ((raw->u.fil.Flags & 1) ? 0 : S_IWUGO) | S_IRUGO | S_IXUGO | S_IFREG;
#else
	inode->i_mode = ((raw->u.fil.Flags & 1) ? 0 : S_IWUGO) | S_IRUGO | S_IFREG;
#endif
	inode->i_nlink = 1;
	inode->i_size = from_l(lsize);
	inode->i_blocks = from_l(psize) >> HFS_SECTOR_SIZE_BITS;
}

typedef int (*iget_func)(struct inode *, struct hfs_brec *, unsigned int, unsigned int, int);

/*
 * See the comment preceeding hfs_notify_change() for a discussion
 * of a kernel change that would make this much cleaner.
 */
static struct inode * common_iget(struct super_block *sb, unsigned int parent,
			int len, const char * name, int type, iget_func fill_in)
{
	struct inode * inode;
	unsigned int ino;
	struct hfs_brec *entry;
	HFS_LWORD mtime, ctime;

	if (!sb || !len || !name || !fill_in || hfs_cat_find(sb, parent, len, name, &entry))
		goto bail3;
	if (RAW->cdrType == HFS_CDR_FIL) {
		ino = from_l(RAW->u.fil.FlNum);
	} else if (RAW->cdrType == HFS_CDR_DIR) {
		ino = from_l(RAW->u.dir.DirID);
	} else {
		printk("hfs_iget: entry is neither file nor directory\n");
		goto bail2;
	}
	if (!(inode = iget(sb, ino | type)))
		goto bail2;
	if ((inode->i_dev == sb->s_dev) && (!inode->i_mode)) {
		memcpy(HFS_I(inode)->key, entry->key, sizeof(HFS_I(inode)->key));
		if (RAW->cdrType==HFS_CDR_FIL) {
			mtime = RAW->u.fil.MdDat;
			ctime = RAW->u.fil.CrDat;
		} else {
			mtime = RAW->u.dir.MdDat;
			ctime = RAW->u.dir.CrDat;
			inode->i_blocks = 0;
		}
		inode->i_rdev = 0;
		inode->i_ctime = m_to_utime(ctime);
		inode->i_atime = inode->i_mtime = m_to_utime(mtime);
		inode->i_uid = HFS_SB(sb)->s_uid;
		inode->i_gid = HFS_SB(sb)->s_gid;
		inode->i_blksize = HFS_SECTOR_SIZE;
		if (fill_in(inode, entry, ino, parent, type))
			goto bail1;
		inode->i_mode &= ~HFS_SB(sb)->s_umask;
	}
	hfs_brec_relse(&entry);
	return inode;

bail1:
	clear_inode(inode);
bail2:
	hfs_brec_relse(&entry);
bail3:
	return NULL;
}

static int complete_cap(struct inode * inode, struct hfs_brec *entry,
	unsigned int ino, unsigned int parent, int type)
{
	if (type == HFS_CAP_FNDR) {
		inode->i_size = sizeof(struct hfs_cap_info);
		inode->i_blocks = 0;
		inode->i_nlink = 1;
		inode->i_mode = ((RAW->u.fil.Flags & 1) ? 0 : S_IWUGO) | S_IRUGO | S_IFREG;
		HFS_I(inode)->u.cap_info.meta = build_cap_info(entry);
		inode->i_op = &hfs_cap_info_inode_operations;
	} else if (RAW->cdrType == HFS_CDR_FIL) {
		unsigned char fork;

		if (type==HFS_CAP_DATA)
			fork = HFS_FK_DATA;
		else if (type==HFS_CAP_RSRC)
			fork = HFS_FK_RSRC;
		else
			return 1;
		read_file_inode(&(HFS_I(inode)->u.file), inode, RAW, fork, ino);
		inode->i_op = &hfs_file_inode_operations;
	} else { /* Directory */
		int valence = from_w(RAW->u.dir.Val);
		if (count_entries(inode->i_sb, ino, valence, &HFS_I(inode)->u.dir.dirs, &HFS_I(inode)->u.dir.files))
			return 1;
		inode->i_size = valence+4;
		inode->i_nlink = 4+HFS_I(inode)->u.dir.dirs;
		inode->i_mode = S_IRWXUGO | S_IFDIR;
		inode->i_op = &hfs_cap_dir_inode_operations;
		HFS_I(inode)->u.dir.parent = parent;
	}
	return 0;
}

static int complete_dbl(struct inode * inode, struct hfs_brec *entry,
	unsigned int ino, unsigned int parent, int type)
{
	if (type == HFS_DBL_HDR) {
		if (RAW->cdrType == HFS_CDR_FIL) {
			read_file_inode(&(HFS_I(inode)->u.dbl_hdr.rsrc), inode, RAW, HFS_FK_RSRC, ino);
			inode->i_size += HFS_HDR_LEN;
#ifdef FAKE_EXEC
			inode->i_mode &= ~S_IXUGO;
#endif
		} else {
			inode->i_size = HFS_HDR_LEN;
			inode->i_mode = ((RAW->u.fil.Flags & 1) ? 0 : S_IWUGO) | S_IRUGO | S_IFREG;
			inode->i_nlink = 1;
		}
		HFS_I(inode)->u.dbl_hdr.meta = build_dbl_hdr(RAW);
		inode->i_op = &hfs_dbl_hdr_inode_operations;
	} else if (RAW->cdrType == HFS_CDR_FIL) {
		read_file_inode(&(HFS_I(inode)->u.file), inode, RAW, HFS_FK_DATA, ino);
		inode->i_op = &hfs_file_inode_operations;
	} else { /* Directory */
		int valence = from_w(RAW->u.dir.Val);
		if (count_entries(inode->i_sb, ino, valence, &HFS_I(inode)->u.dir.dirs, &HFS_I(inode)->u.dir.files))
			return 1;
		
		inode->i_size = 2+2*valence;
		inode->i_nlink = 2+HFS_I(inode)->u.dir.dirs;
		inode->i_mode = S_IRWXUGO | S_IFDIR;
		inode->i_op = &hfs_dbl_dir_inode_operations;
		HFS_I(inode)->u.dir.parent = parent;
	}
	return 0;
}

struct inode * hfs_cap_iget(struct super_block *sb, unsigned int parent,
	int len, const char * name, int type)
{
	return common_iget(sb, parent, len, name, type, complete_cap);
}

struct inode * hfs_dbl_iget(struct super_block *sb, unsigned int parent,
	int len, const char * name, int type)
{
	return common_iget(sb, parent, len, name, type, complete_dbl);
}

static void write_file_inode(struct hfs_file *file, struct inode *inode,
	struct hfs_inode *raw)
{
	if (file->magic == HFS_FILE_MAGIC) {
		if (file->fork == HFS_FK_DATA) {
			raw->u.fil.LgLen = to_l(inode->i_size);
			raw->u.fil.PyLen = to_l(inode->i_blocks << HFS_SECTOR_SIZE_BITS);
			hfs_extent_out(file, &raw->u.fil.ExtRec);
		} else {
			raw->u.fil.RLgLen = to_l(inode->i_size);
			raw->u.fil.RPyLen = to_l(inode->i_blocks << HFS_SECTOR_SIZE_BITS);
			hfs_extent_out(file, &raw->u.fil.RExtRec);
		}
	} else {
		printk("hfs_write_inode: corrupted hfs_file structure!\n");
	}
}

static void common_write_inode(struct inode * inode, struct hfs_brec *entry)
{
	HFS_LWORD mtime, ctime;

	mtime = u_to_mtime(inode->i_mtime);
	ctime = u_to_mtime(inode->i_ctime);
	if (S_ISREG(inode->i_mode)) {
		RAW->u.fil.Flags &= ~1;
		RAW->u.fil.Flags |= (inode->i_mode & S_IWUSR ? 0 : 1);
		RAW->u.fil.MdDat = mtime;
		RAW->u.fil.CrDat = ctime;
	} else if (S_ISDIR(inode->i_mode)) {
		RAW->u.dir.Val = to_w(HFS_I(inode)->u.dir.dirs + 
				      HFS_I(inode)->u.dir.files);
		RAW->u.dir.MdDat = mtime;
		RAW->u.dir.CrDat = ctime;
	}
}

void hfs_cap_write_inode(struct inode * inode)
{
	struct hfs_brec *entry;

	if (!hfs_bfind(HFS_SB(inode->i_sb)->s_cat_tree, (struct hfs_rec_key *)(HFS_I(inode)->key), &entry, HFS_BFIND_WRITE)) {
		common_write_inode(inode, entry);
		if (S_ISREG(inode->i_mode))
			write_file_inode(&HFS_I(inode)->u.file, inode, RAW);
		hfs_brec_relse(&entry);
	} else {
		printk("Unable to read inode %ld\n", inode->i_ino);
	}
	inode->i_dirt = 0;
}

void hfs_dbl_write_inode(struct inode * inode)
{
	struct hfs_brec *entry;

	if (!hfs_bfind(HFS_SB(inode->i_sb)->s_cat_tree, (struct hfs_rec_key *)(HFS_I(inode)->key), &entry, HFS_BFIND_WRITE)) {
		common_write_inode(inode, entry);
		if (S_ISREG(inode->i_mode))
			if (HFS_ITYPE(inode->i_ino) == HFS_DBL_DATA)
				write_file_inode(&HFS_I(inode)->u.file, inode, RAW);
			else
				write_file_inode(&HFS_I(inode)->u.dbl_hdr.rsrc, inode, RAW);
		hfs_brec_relse(&entry);
	} else {
		printk("Unable to read inode %ld\n", inode->i_ino);
	}
	inode->i_dirt = 0;
}

#define	HFS_VALID_MODE_BITS	(S_IFREG | S_IFDIR | S_IRWXU | S_IRWXG | S_IRWXO)

/*
 * Copied from fs/msdos/inode.c by Werner Almesberger
 *
 * This function takes the requested change to the inode and
 * makes the closest possible change.
 * 
 * XXX It would be really nice if instead of calling iget(), there
 * was a way to get the inode if it were in core, or NULL if not.
 * Alternatively we could keep track of all the in-core inodes, by
 * tracking their creation and deletion in hfs_*_iget() and
 * hfs_*_put_inode(). However, this would mean duplicating the
 * inode cache used by iget(), so is not practical.
 * This comment also applies to the way hfs_*_iget is written.
 */
int hfs_notify_change(struct inode * inode,struct iattr * attr)
{
	int error, old_dirt;
	unsigned int type, ino;
	struct inode *tmpi;
	umode_t owner_bits;


	error = inode_change_ok(inode, attr);
	if (error)
		return error;

	/* XXX until we have a valid i_op->truncate() */
	if ((attr->ia_valid & ATTR_SIZE) &&
	    (attr->ia_size != inode->i_size))
		return -EPERM;

	if (((attr->ia_valid & ATTR_UID) && 
	     (attr->ia_uid != HFS_SB(inode->i_sb)->s_uid)) ||
	    ((attr->ia_valid & ATTR_GID) && 
	     (attr->ia_gid != HFS_SB(inode->i_sb)->s_gid)) ||
	    ((attr->ia_valid & ATTR_MODE) &&
	     (attr->ia_mode & ~HFS_VALID_MODE_BITS)))
		error = -EPERM;

	/* should we have a quiet mount flag tested here like msdos? */
	if (error)
		return error;
	
	/* make the mode bits legal for the file's owner: */
	if (S_ISDIR(inode->i_mode)) {
		owner_bits = S_IRWXU;
	} else {
#ifdef FAKE_EXEC
		owner_bits = S_IXUSR | S_IRUSR | (attr->ia_mode & S_IWUSR);
#else
		owner_bits = S_IRUSR | (attr->ia_mode & S_IWUSR);
#endif
	}
	owner_bits &= ~HFS_SB(inode->i_sb)->s_umask;

	attr->ia_mode = ((attr->ia_mode & S_IFMT) | owner_bits |
			 (owner_bits >> 3) | (owner_bits >> 6)) &
			~HFS_SB(inode->i_sb)->s_umask;

	inode_setattr(inode, attr);

	/* We must change all in-core inodes correspong to this file. */
	ino = HFS_IFID(inode->i_ino);
	for (type=0; type < (1<<HFS_ITYPE_BITS); ++type) {
		if ((type<<HFS_ITYPE_SHIFT) == HFS_ITYPE(inode->i_ino))
			continue;
		tmpi = iget(inode->i_sb, ino | (type<<HFS_ITYPE_SHIFT));
		if ((tmpi->i_dev != inode->i_dev) ||
		    (tmpi->i_ino != (ino | (type<<HFS_ITYPE_SHIFT)))) {
			iput(tmpi);
		} else if (!tmpi->i_mode) {
			clear_inode(tmpi);
		} else {
			old_dirt = tmpi->i_dirt;
			inode_setattr(tmpi, attr);
			tmpi->i_dirt = old_dirt;
			iput(tmpi);
		}
	}

	return 0;
}
