/*
 *  linux/fs/hfs/super.c
 *
 *  Copyright (C) 1995  Paul H. Hargrove
 *  This file may be distributed under the terms of the GPL
 *
 *  Based on the minix file system code, (C) 1991, 1992 by Linus Torvalds
 *
 *  Code to handle the Mac partition table based on
 *  a patch contributed by Holger Schemel.
 */

#include <linux/config.h>
#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include "hfs_fs.h"
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/locks.h>
#include <linux/stddef.h>

/* The Mac Driver Descriptor Block */
struct hfs_drvr_desc {
	HFS_WORD	ddSig		DONT_ALIGN;
	/* a bunch more stuff we don't need */
};

/* The Mac partiton map */
struct hfs_partition_map {
	HFS_WORD	pmSig		DONT_ALIGN;
	HFS_WORD 	reSigPad	DONT_ALIGN;
	HFS_LWORD	pmMapBlkCnt	DONT_ALIGN;
	HFS_LWORD	pmPyPartStart	DONT_ALIGN;
	HFS_LWORD	pmPartBlkCnt	DONT_ALIGN;
	HFS_BYTE	pmPartName[32]	DONT_ALIGN;
	HFS_BYTE 	pmPartType[32]	DONT_ALIGN;
	/* a bunch more stuff we don't need */
};

/* The HFS superblock */
struct hfs_super {
	HFS_WORD	drSigWord	DONT_ALIGN;
	HFS_LWORD	drCrDate	DONT_ALIGN;
	HFS_LWORD	drLastMod	DONT_ALIGN;
	HFS_WORD	drAtrb		DONT_ALIGN;
	HFS_WORD	drNmFls		DONT_ALIGN;
	HFS_WORD	drVBMSt		DONT_ALIGN;
	HFS_WORD	drAllocPtr	DONT_ALIGN;
	HFS_WORD	drNmAlBlks	DONT_ALIGN;
	HFS_LWORD	drAlBlkSiz	DONT_ALIGN;
	HFS_LWORD	drClpSiz	DONT_ALIGN;
	HFS_WORD	drAlBlSt	DONT_ALIGN;
	HFS_LWORD	drNxtCNID	DONT_ALIGN;
	HFS_WORD	drFreeBks	DONT_ALIGN;
	HFS_BYTE	drVNlen		DONT_ALIGN;
	HFS_BYTE	drVN[27]	DONT_ALIGN;
	HFS_LWORD	drVolBkUp	DONT_ALIGN;
	HFS_WORD	drVSeqNum	DONT_ALIGN;
	HFS_LWORD	drWrCnt		DONT_ALIGN;
	HFS_LWORD	drXTClpSiz	DONT_ALIGN;
	HFS_LWORD	drCTClpSiz	DONT_ALIGN;
	HFS_WORD	drNmRtDirs	DONT_ALIGN;
	HFS_LWORD	drFilCnt	DONT_ALIGN;
	HFS_LWORD	drDirCnt	DONT_ALIGN;
	HFS_BYTE	drFndrInfo[32]	DONT_ALIGN;
	HFS_WORD	drVCSize	DONT_ALIGN;
	HFS_WORD	drVCBMSize	DONT_ALIGN;
	HFS_WORD	drCtlCSize	DONT_ALIGN;
	HFS_LWORD	drXTFlSize	DONT_ALIGN;
	struct hfs_raw_extent
			drXTExtRec	DONT_ALIGN;
	HFS_LWORD	drCTFlSize	DONT_ALIGN;
	struct hfs_raw_extent
			drCTExtRec	DONT_ALIGN;
};

/* called with super block locked */
static void do_put_super(struct super_block *sb)
{
	int lcv;

	sb->s_dev = 0;
	if (HFS_SB(sb)->s_ext_tree)
		hfs_btree_free(HFS_SB(sb)->s_ext_tree);
	if (HFS_SB(sb)->s_cat_tree)
		hfs_btree_free(HFS_SB(sb)->s_cat_tree);
	if (HFS_SB(sb)->s_sbbh)
		brelse(HFS_SB(sb)->s_sbbh);
	for (lcv = 0; HFS_SB(sb)->s_bitmap[lcv] && (lcv < 16); ++lcv)
		brelse(HFS_SB(sb)->s_bitmap[lcv]);

	MOD_DEC_USE_COUNT;
	return;
}

static void hfs_put_super(struct super_block *sb)
{
	lock_super(sb);
	do_put_super(sb);
	unlock_super(sb);
	return;
}

#ifndef LINUX_1_2
static void hfs_statfs(struct super_block *sb, struct statfs *buf, int len)
{
	struct statfs tmp;

	tmp.f_type = HFS_SUPER_MAGIC;
	tmp.f_bsize = HFS_SECTOR_SIZE;
	tmp.f_blocks = HFS_SB(sb)->s_alloc_blksz * HFS_SB(sb)->s_fs_ablocks;
	tmp.f_bfree = HFS_SB(sb)->s_alloc_blksz * HFS_SB(sb)->s_free_ablocks;
	tmp.f_bavail = tmp.f_bfree;
	tmp.f_files = HFS_SB(sb)->s_file_count + HFS_SB(sb)->s_dir_count;
	tmp.f_ffree = ~(HFS_SB(sb)->s_next_id);
	tmp.f_namelen = HFS_NAMELEN;

	memcpy_tofs(buf, &tmp, len);
}
#else /* LINUX_1_2 */
static void hfs_statfs(struct super_block *sb, struct statfs *buf)
{
	unsigned long total = HFS_SB(sb)->s_alloc_blksz * HFS_SB(sb)->s_fs_ablocks;
	unsigned long avail = HFS_SB(sb)->s_alloc_blksz * HFS_SB(sb)->s_free_ablocks;

	put_fs_long(HFS_SUPER_MAGIC, &buf->f_type);
	put_fs_long(HFS_SECTOR_SIZE, &buf->f_bsize);
	put_fs_long(total, &buf->f_blocks);
	put_fs_long(avail, &buf->f_bfree);
	put_fs_long(avail, &buf->f_bavail);
	put_fs_long(HFS_SB(sb)->s_file_count + HFS_SB(sb)->s_dir_count,
			&buf->f_files);
	put_fs_long(~(HFS_SB(sb)->s_next_id), &buf->f_ffree);
	put_fs_long(HFS_NAMELEN, &buf->f_namelen);
	/* I don't know what value to put in buf->f_fsid either! */
}
#endif /* LINUX_1_2 */

static void hfs_cap_put_inode(struct inode * inode)
{
	if ((HFS_ITYPE(inode->i_ino)==HFS_CAP_FNDR) && HFS_I(inode)->u.cap_info.meta) {
		/* since we can't hold on to the pointer to the mapped data */
		/* we'll need to re-read it each time we fetch the inode */
		/* this could be improved to be only upon open(), but is */
		/* still better than getting it from disk on each read() */
		kfree_s(HFS_I(inode)->u.cap_info.meta, sizeof(struct hfs_cap_info));
		clear_inode(inode);
	}
}

static void hfs_dbl_put_inode(struct inode * inode)
{
	if ((HFS_ITYPE(inode->i_ino)==HFS_DBL_HDR) && HFS_I(inode)->u.dbl_hdr.meta) {
		/* since we can't hold on to the pointer to the mapped data */
		/* we'll need to re-read it each time we fetch the inode */
		/* this could be improved to be only upon open(), but is */
		/* still better than getting it from disk on each read() */
		kfree_s(HFS_I(inode)->u.dbl_hdr.meta, HFS_HDR_LEN);
		clear_inode(inode);
	}
}

static struct super_operations hfs_cap_super_operations = { 
	NULL,			/* read_inode */
	hfs_notify_change,	/* notify_change */
	hfs_cap_write_inode,	/* write_inode */
	hfs_cap_put_inode,	/* put_inode */
	hfs_put_super,		/* put_super */
	NULL,			/* write_super */
	hfs_statfs,		/* statfs */
	NULL			/* remount_fs */
};

static struct super_operations hfs_dbl_super_operations = { 
	NULL,			/* read_inode */
	hfs_notify_change,	/* notify_change */
	hfs_dbl_write_inode,	/* write_inode */
	hfs_dbl_put_inode,	/* put_inode */
	hfs_put_super,		/* put_super */
	NULL,			/* write_super */
	hfs_statfs,		/* statfs */
	NULL			/* remount_fs */
};

/* 
 * adapted from linux/fs/msdos/inode.c written 1992,93 by Werner Almesberger
 */
static int parse_options(char *options, uid_t *uid, gid_t *gid, int *umask,
	char *names, char *fork, int *part)
{
	char *this_char,*value;

	*uid = current->uid;
	*gid = current->gid;
	*umask = current->fs->umask;
	*names = 'c';
	*fork = 'c';
	*part = 0;
	if (!options) return 1;
	for (this_char = strtok(options,","); this_char; this_char = strtok(NULL,",")) {
		if ((value = strchr(this_char,'=')) != NULL)
			*value++ = 0;
		if (!strcmp(this_char,"uid")) {
			if (!value || !*value)
				return 0;
			*uid = simple_strtoul(value,&value,0);
			if (*value)
				return 0;
		} else if (!strcmp(this_char,"gid")) {
			if (!value || !*value)
				return 0;
			*gid = simple_strtoul(value,&value,0);
			if (*value)
				return 0;
		} else if (!strcmp(this_char,"umask")) {
			if (!value || !*value)
				return 0;
			*umask = simple_strtoul(value,&value,8);
			if (*value)
				return 0;
		} else if (!strcmp(this_char,"part")) {
			if (!value || !*value)
				return 0;
			*part = simple_strtoul(value,&value,8);
			if (*value)
				return 0;
		} else if (!strcmp(this_char,"names") && value) {
			if (value[0] && !value[1] && strchr("ac78tl",*value))
				*names = *value;
			else if (!strcmp(value,"alpha")) *names = 'a';
			else if (!strcmp(value,"cap")) *names = 'c';
			else if (!strcmp(value,"7bit")) *names = '7';
			else if (!strcmp(value,"8bit")) *names = '8';
			else if (!strcmp(value,"trivial")) *names = 't';
			else if (!strcmp(value,"latin")) *names = 'l';
			else return 0;
		} else if (!strcmp(this_char,"fork") && value) {
			if (value[0] && !value[1] && strchr("cds",*value))
				*fork = *value;
			else if (!strcmp(value,"cap")) *fork = 'c';
			else if (!strcmp(value,"double")) *fork = 'd';
			else if (!strcmp(value,"single")) *fork = 's';
			else return 0;
		} else
			return 0;
	}
	return 1;
}

struct super_block *hfs_read_super(struct super_block *s,void *data, int silent)
{
	struct hfs_super *hs;
	int lcv, part, umask, dev=s->s_dev;
	uid_t uid;
	gid_t gid;
	unsigned int bs, block;
	unsigned int partition_start = 0;
	char names, fork;
	const char *err;
	struct buffer_head *bh, **bmbh;
	struct hfs_partition_map *pm;

	/* in case someone tries to unload the module while we wait on I/O */
	MOD_INC_USE_COUNT;

	if (!parse_options((char *)data, &uid, &gid, &umask, &names, &fork, &part)) {
		err = "unable to parse mount options";
		MOD_DEC_USE_COUNT;
		goto bail2;
	}

	lock_super(s);

	HFS_SB(s)->s_ext_tree = HFS_SB(s)->s_cat_tree = NULL;
	HFS_SB(s)->s_uid = uid;
	HFS_SB(s)->s_gid = gid;
	HFS_SB(s)->s_umask = umask;
	if (names=='c')	{
		HFS_SB(s)->s_nameout = hfs_cap2mac;
		HFS_SB(s)->s_namein = hfs_mac2cap;
	} else if (names=='t') {
		HFS_SB(s)->s_nameout = hfs_triv2mac;
		HFS_SB(s)->s_namein = hfs_mac2triv;
	} else if (names=='7') {
		HFS_SB(s)->s_namein = hfs_mac2seven;
		HFS_SB(s)->s_nameout = hfs_apl2mac;
	} else if (names=='8') {
		HFS_SB(s)->s_namein = hfs_mac2eight;
		HFS_SB(s)->s_nameout = hfs_apl2mac;
	} else if (names=='l') {
		HFS_SB(s)->s_nameout = hfs_latin2mac;
		HFS_SB(s)->s_namein = hfs_mac2latin;
	} else {/* names=='a' */
		HFS_SB(s)->s_namein = hfs_mac2alpha;
		HFS_SB(s)->s_nameout = hfs_apl2mac;
	}

	set_blocksize(dev,HFS_SECTOR_SIZE);

	if (!(bh = bread(dev, HFS_PMAP_BLK, HFS_SECTOR_SIZE))) {
		err = "unable to read partition map block";
		goto bail1;
	}

	pm = (struct hfs_partition_map *)(bh->b_data);
	if (from_w(pm->pmSig) == HFS_OLD_PMAP_MAGIC) {
		printk("hfs_read_super: This disk has an unsupported old-style partition map.\n");
		err = "If you'd like to see support for this please contact the author.";
		goto bail0;
	} else if (from_w(pm->pmSig) == HFS_NEW_PMAP_MAGIC) {
		int i, pmap_entries;

		pmap_entries = from_l(pm->pmMapBlkCnt);
		for (i=0; i<pmap_entries && !partition_start; ++i) {
			/* read the entry if we don't have it yet */
			if (i != 0) {
				brelse(bh);
				if (!(bh = bread(dev, HFS_PMAP_BLK + i, HFS_SECTOR_SIZE))) {
					err = "unable to read partition map block";
					goto bail1;
				}
				pm = (struct hfs_partition_map *)(bh->b_data);
				if (from_w(pm->pmSig) != HFS_NEW_PMAP_MAGIC) {
					err = "invalid entry in partition map";
					goto bail0;
				}
			}
			/* look for an HFS partition */
			if (!memcmp(pm->pmPartType, "Apple_HFS", 9) && ((--part) <= 0)) {
				partition_start = from_l(pm->pmPyPartStart);
			}
		}
		if (part > 0) {
			err = "unable to locate the requested parition";
			goto bail0;
		}
		if (!partition_start) {
			err = "no HFS parition was found";
			goto bail0;
		}
	}
	brelse(bh);
				
	/* if we get here we are on unpartitioned media or found an HFS partition */
	if (!(HFS_SB(s)->s_sbbh = bread(dev, HFS_MDB_BLK + partition_start, HFS_SECTOR_SIZE))) {
		err = "unable to read superblock";
		goto bail1;
	}

	HFS_SB(s)->s_hs = hs = (struct hfs_super *)HFS_SB(s)->s_sbbh->b_data;

	if ((s->s_magic = from_w(hs->drSigWord)) != HFS_SUPER_MAGIC) {
		if (!silent)
			printk("VFS: Can't find a HFS filesystem on dev 0x%04x.\n",
				   dev);
		err = NULL;
		goto bail1;
	}

	bs = from_l(hs->drAlBlkSiz);
	if (bs > HFS_USHRT_MAX || (bs < HFS_SECTOR_SIZE) || (bs & (HFS_SECTOR_SIZE-1))) {
		err = "bad allocation block size";
		goto bail1;
	}
	s->s_blocksize_bits = HFS_SECTOR_SIZE_BITS;
	HFS_SB(s)->s_alloc_blksz = bs >> s->s_blocksize_bits;

	/* These parameters are read from the super block, and never written */
	HFS_SB(s)->s_fs_ablocks = from_w(hs->drNmAlBlks);
	HFS_SB(s)->s_fs_start = from_w(hs->drAlBlSt) + partition_start;
	HFS_SB(s)->s_clump_size = from_l(hs->drClpSiz);

	/* These parameters are read from and written to the super block */
	HFS_SB(s)->s_free_ablocks = from_w(hs->drFreeBks);
	HFS_SB(s)->s_next_id = from_l(hs->drNxtCNID);
	HFS_SB(s)->s_write_count = from_l(hs->drWrCnt);
	HFS_SB(s)->s_root_files = from_w(hs->drNmFls);
	HFS_SB(s)->s_root_dirs = from_w(hs->drNmRtDirs);
	HFS_SB(s)->s_file_count = from_l(hs->drFilCnt);
	HFS_SB(s)->s_dir_count = from_l(hs->drDirCnt);
	
	/* read in the bitmap */
	memset(HFS_SB(s)->s_bitmap, 0, sizeof(HFS_SB(s)->s_bitmap));
	block = from_w(hs->drVBMSt) + partition_start;
	bmbh = HFS_SB(s)->s_bitmap;
	lcv = (HFS_SB(s)->s_fs_ablocks + 4095) / 4096;
	for ( ; lcv; --lcv, ++bmbh, ++block) {
		if (!(*bmbh = bread(dev, block, HFS_SECTOR_SIZE))) {
			err = "unable to read bitmap";
			goto bail1;
		}
	}

	if (HFS_ITYPE(HFS_SB(s)->s_next_id) != 0) {
		err = "too many files";
		goto bail1;
	}

	if (!(HFS_SB(s)->s_ext_tree = hfs_btree_init(s, HFS_EXT_FID, &hs->drXTExtRec, hs->drXTFlSize, hs->drXTClpSiz)) ||
	    !(HFS_SB(s)->s_cat_tree = hfs_btree_init(s, HFS_CAT_FID, &hs->drCTExtRec, hs->drCTFlSize, hs->drCTClpSiz))) {
		err = "unable to initialize data structures";
		goto bail1;
	}

	s->s_blocksize = HFS_SECTOR_SIZE;
	switch (fork) {
	case 's':
		err = "AppleSingle not yet implemented";
		goto bail1;
		/* break; */
	
	case 'd':
		s->s_op = &hfs_dbl_super_operations;
		s->s_mounted = hfs_dbl_iget(s, 1, hs->drVNlen, hs->drVN, HFS_DBL_DIR);
		break;
	
	case 'c':
		s->s_op = &hfs_cap_super_operations;
		s->s_mounted = hfs_cap_iget(s, 1, hs->drVNlen, hs->drVN, HFS_CAP_NDIR);
	}
	if (!s->s_mounted) {
		err = "get root inode failed";
		goto bail1;
	}
	unlock_super(s);
	return s;

bail0:
	brelse(bh);
bail1:
	do_put_super(s);
	unlock_super(s);
bail2:
	s->s_dev = 0;
	if (err)
		printk("hfs_read_super: %s.\n", err);
	return NULL;	
}

#if defined(DEBUG_BRECS) || defined(DEBUG_ALL)
int hfs_bcount = 0;
#endif

#ifdef MODULE
char kernel_version[] = UTS_RELEASE;

static struct file_system_type hfs_fs = {hfs_read_super,"hfs",1,NULL};

int init_module(void) {
#ifdef PARANOID
	/* too bad sizeof() is not allowed in an '#if' */
	/* GCC optimizes the printk() away, but the string */
	/* still takes up 32 bytes of "valuable" space!! */
	if ((sizeof(struct hfs_sb_info)>sizeof(struct ext2_sb_info)) ||
	    (sizeof(struct hfs_inode_info)>sizeof(struct ext2_inode_info))) {
		printk("Filesystem structures too big!!\n");
		return -ENOMEM; /* well sort of */
	}
#endif
/*	register_symtab(NULL);	*/
	register_filesystem(&hfs_fs);
	return 0;
}

void cleanup_module(void) {
#if defined(DEBUF_BRECS) || defined(DEBUG_ALL)
	printk("hfs_bcount = %d\n", hfs_bcount);
#endif
	unregister_filesystem(&hfs_fs);
}
#endif
