/*
 *  linux/fs/hfs/trans.c
 *
 *  Copyright (C) 1995  Paul H. Hargrove
 *  This file may be distributed under the terms of the GPL
 *
 *  Latin-1 translation based on a patch contributed by Holger Schemel.
 */

#include <linux/config.h>
#ifdef MODULE
#include <linux/module.h>
#endif

#include "hfs_fs.h"

static __inline__ const unsigned char dehex(char c) {
	if ((c>='0')&&(c<='9'))
		return c-'0';
	if ((c>='a')&&(c<='f'))
		return c-'a'+10;
	if ((c>='A')&&(c<='F'))
		return c-'A'+10;
	return 0xff;
}

static char hex[16] = 
	{'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};

static unsigned char latin2mac_map[128]; /* initially all zero */

static unsigned char mac2latin_map[128] = {
	0xC4, 0xC5, 0xC7, 0xC9, 0xD1, 0xD6, 0xDC, 0xE1,
	0xE0, 0xE2, 0xE4, 0xE3, 0xE5, 0xE7, 0xE9, 0xE8,
	0xEA, 0xEB, 0xED, 0xEC, 0xEE, 0xEF, 0xF1, 0xF3,
	0xF2, 0xF4, 0xF6, 0xF5, 0xFA, 0xF9, 0xFB, 0xFC,
	0x00, 0xB0, 0xA2, 0xA3, 0xA7, 0xB7, 0xB6, 0xDF,
	0xAE, 0xA9, 0x00, 0xB4, 0xA8, 0x00, 0xC6, 0xD8,
	0x00, 0xB1, 0x00, 0x00, 0xA5, 0xB5, 0xF0, 0x00, 
	0x00, 0x00, 0x00, 0xAA, 0xBA, 0x00, 0xE6, 0xF8,
	0xBF, 0xA1, 0xAC, 0x00, 0x00, 0x00, 0x00, 0xAB,
	0xBB, 0x00, 0xA0, 0xC0, 0xC3, 0xD5, 0x00, 0x00, 
	0xAD, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF7, 0x00, 
	0xFF, 0x00, 0x00, 0xA4, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0xB8, 0x00, 0x00, 0xC2, 0xCA, 0xC1,
	0xCB, 0xC8, 0xCD, 0xCE, 0xCF, 0xCC, 0xD3, 0xD4,
	0x00, 0xD2, 0xDA, 0xDB, 0xD9, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

int hfs_mac2cap(char *out, const unsigned char *in) {
	unsigned char c;
	int len = *in++;
	int count = 0;

	if (!out || !in)
		return -1;

	while (len--) {
		c = *in++;
		if ((c<32) || (c=='/') || (c>126)) {
			*out++ = ':';
			*out++ = hex[(c>>4) & 0xf];
			*out++ = hex[c & 0xf];
			count += 3;
		} else {
			*out++ = c;
			count++;
		}
	}
	return count;
}

int hfs_mac2eight(char *out, const unsigned char *in) {
	unsigned char c;
	int len = *in++;
	int count = 0;

	if (!out || !in)
		return -1;

	while (len--) {
		c = *in++;
		if (!c || (c=='/') || (c=='%')) {
			*out++ = '%';
			*out++ = hex[(c>>4) & 0xf];
			*out++ = hex[c & 0xf];
			count += 3;
		} else {
			*out++ = c;
			count++;
		}
	}
	return count;
}

int hfs_mac2seven(char *out, const unsigned char *in) {
	unsigned char c;
	int len = *in++;
	int count = 0;

	if (!out || !in)
		return -1;

	while (len--) {
		c = *in++;
		if (!c || (c=='/') || (c=='%') || (c&0x80)) {
			*out++ = '%';
			*out++ = hex[(c>>4) & 0xf];
			*out++ = hex[c & 0xf];
			count += 3;
		} else {
			*out++ = c;
			count++;
		}
	}
	return count;
}

int hfs_mac2alpha(char *out, const unsigned char *in) {
	unsigned char c;
	int len = *in++;
	int count = 0;
	const unsigned char *lp;	/* last period */

	if (!out || !in)
		return -1;

	/* strrchr() would be good here, but 'in' is not null-terminated */
	for(lp=in+len-1; (lp>=in)&&(*lp!='.'); --lp) ;
	++lp;

	while (len--) {
		c = *in++;
		if ((in==lp) || ((c>='0')&&(c<='9')) || ((c>='A')&&(c<='Z')) ||
				((c>='a')&&(c<='z')) || (c=='_')) {
			*out++ = c;
			count++;
		} else {
			*out++ = '%';
			*out++ = hex[(c>>4) & 0xf];
			*out++ = hex[c & 0xf];
			count += 3;
		}
	}
	return count;
}

int hfs_mac2triv(char *out, const unsigned char *in) {
	unsigned char c;
	int len = *in++;
	int count = 0;

	if (!out || !in)
		return -1;

	while (len--) {
		c = *in++;
		if (c=='/') {
			*out++ = ':';
		} else {
			*out++ = c;
		}
		count++;
	}
	return count;
}

int hfs_mac2latin(char *out, const unsigned char *in) {
	unsigned char c;
	int len = *in++;
	int count = 0;

	if (!out || !in)
		return -1;

	while (len--) {
		c = *in++;

		if (c>127 && mac2latin_map[c-128]) {
			*out++ = mac2latin_map[c-128];
			count++;
		} else if (c<32 || c>126 || (c=='/') || (c=='%')) {
			*out++ = '%';
			*out++ = hex[(c>>4) & 0xf];
			*out++ = hex[c & 0xf];
			count += 3;
		} else {
			*out++ =  c;
			count++;
		}
	}
	return count;
}

int hfs_cap2mac(char *out, int len, const char *in) {
	int hi, lo, count = 0;
	unsigned char c;

	if (!in || !out)
		return -1;

	while (len-- && (count <= HFS_NAMELEN)) {
		c = *in++;
		count++;
		if (c!=':') {
			*out++ = c;
		} else if ((len<2) ||
			   ((hi=dehex(in[0])) & 0xf0) ||
			   ((lo=dehex(in[1])) & 0xf0)) {
			*out++ = '|';
		} else {
			*out++ = (hi << 4) | lo;
			len -= 2;
			in += 2;
		}
	}
	return count;
}

int hfs_apl2mac(char *out, int len, const char *in) {
	int hi, lo, count = 0;
	unsigned char c;

	if (!in || !out)
		return -1;

	while (len-- && (count <= HFS_NAMELEN)) {
		c = *in++;
		count++;
		if (c==':') {
			return -1;
		} else if (c!='%') {
			*out++ = c;
		} else if ((len<2) ||
			   ((hi=dehex(in[0])) & 0xf0) ||
			   ((lo=dehex(in[1])) & 0xf0)) {
			*out++ = '%';
		} else {
			*out++ = (hi << 4) | lo;
			len -= 2;
			in += 2;
		}
	}
	return count;
}

int hfs_triv2mac(char *out, int len, const char *in) {
	unsigned char c;
	int count = 0;

	if (!in || !out)
		return -1;

	while (len-- && (count <= HFS_NAMELEN)) {
		c = *in++;
		if (c==':')
			*out++ = '/';
		else
			*out++ = c;
		count++;
	}
	return count;
}

int hfs_latin2mac(char *out, int len, const char *in)
{
	int hi, lo, count = 0;
	unsigned char c;
	static int map_initialized = 0;

	if (!map_initialized) {
		int i;

		/* build the inverse mapping at run time */
		for(i = 0; i < 128; i++) {
			if ((c = mac2latin_map[i]))
				latin2mac_map[(int)c - 128] = i + 128;
		}
		map_initialized = 1;
	}

	while (len-- && (count <= HFS_NAMELEN)) {
		c = *in++;
		count++;

		if (c != '%') {
			if (c<128 || !(*out = latin2mac_map[c-128]))
				*out = c;
			out++;
		} else if ((len<2) ||
			   ((hi=dehex(in[0])) & 0xf0) ||
			   ((lo=dehex(in[1])) & 0xf0)) {
			*out++ = '%';
		} else {
			*out++ = (hi << 4) | lo;
			len -= 2;
			in += 2;
		}
	}
	return count;
}
